# filepath: /workspaces/peptacular/tests/test_proforma_funcs/add_mods.py
import unittest

import peptacular as pt


class TestAddMods(unittest.TestCase):
    """
    TESTS FOR: add_labile_mods
    """

    def test_add_labile_mods_replace_inplace_true(self):
        annotation = pt.ProFormaAnnotation.parse("{Glycan}PEPTIDE")
        result = annotation.set_labile_mods("HexNAc", inplace=True)

        self.assertIs(result, annotation)  # Should return same object
        self.assertEqual(annotation.serialize(), "{HexNAc}PEPTIDE")

    def test_add_labile_mods_replace_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("{Glycan}PEPTIDE")
        result = annotation.set_labile_mods("HexNAc", inplace=False)

        self.assertIsNot(result, annotation)  # Should return different object
        self.assertEqual(result.serialize(), "{HexNAc}PEPTIDE")
        # Original should be unchanged
        self.assertEqual(annotation.serialize(), "{Glycan}PEPTIDE")

    def test_add_labile_mods_append_to_existing(self):
        annotation = pt.ProFormaAnnotation.parse("{Glycan}PEPTIDE")
        annotation.append_labile_mod("HexNAc", inplace=True)

        self.assertEqual(annotation.serialize(), "{Glycan}{HexNAc}PEPTIDE")

    def test_add_labile_mods_append_to_empty(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.append_labile_mod("Glycan", inplace=True)

        self.assertEqual(annotation.serialize(), "{Glycan}PEPTIDE")

    def test_add_labile_mods_list_input(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.set_labile_mods(["Glycan", "HexNAc"], inplace=True)

        self.assertEqual(annotation.serialize(), "{Glycan}{HexNAc}PEPTIDE")

    def test_add_labile_mods_mod_object_input(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        mod = pt.Mod("Glycan", 1)
        annotation.set_labile_mods(mod, inplace=True)

        self.assertEqual(annotation.serialize(), "{Glycan}PEPTIDE")

    def test_add_labile_mods_none_input_no_append(self):
        annotation = pt.ProFormaAnnotation.parse("{Glycan}PEPTIDE")
        annotation.set_labile_mods(None, inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE")

    """
    TESTS FOR: add_unknown_mods
    """

    def test_add_unknown_mods_replace_inplace_true(self):
        annotation = pt.ProFormaAnnotation.parse("[Unknown]?PEPTIDE")
        result = annotation.set_unknown_mods("Mystery", inplace=True)

        self.assertIs(result, annotation)
        self.assertEqual(annotation.serialize(), "[Mystery]?PEPTIDE")

    def test_add_unknown_mods_replace_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("[Unknown]?PEPTIDE")
        result = annotation.set_unknown_mods("Mystery", inplace=False)

        self.assertIsNot(result, annotation)
        self.assertEqual(result.serialize(), "[Mystery]?PEPTIDE")
        # Original should be unchanged
        self.assertEqual(annotation.serialize(), "[Unknown]?PEPTIDE")

    def test_add_unknown_mods_append_to_existing(self):
        annotation = pt.ProFormaAnnotation.parse("[Unknown]?PEPTIDE")
        annotation.append_unknown_mod("Mystery", inplace=True)

        self.assertEqual(annotation.serialize(), "[Unknown][Mystery]?PEPTIDE")

    def test_add_unknown_mods_append_to_empty(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.append_unknown_mod("Unknown", inplace=True)

        self.assertEqual(annotation.serialize(), "[Unknown]?PEPTIDE")

    def test_add_unknown_mods_list_input(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.set_unknown_mods(["Unknown1", "Unknown2"], inplace=True)

        self.assertEqual(annotation.serialize(), "[Unknown1][Unknown2]?PEPTIDE")

    def test_add_unknown_mods_none_input_no_append(self):
        annotation = pt.ProFormaAnnotation.parse("[Unknown]?PEPTIDE")
        annotation.set_unknown_mods(None, inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE")

    """
    TESTS FOR: add_nterm_mods
    """

    def test_add_nterm_mods_replace_inplace_true(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        result = annotation.set_nterm_mods("Formyl", inplace=True)

        self.assertIs(result, annotation)
        self.assertEqual(annotation.serialize(), "[Formyl]-PEPTIDE")

    def test_add_nterm_mods_replace_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        result = annotation.set_nterm_mods("Formyl", inplace=False)

        self.assertIsNot(result, annotation)
        self.assertEqual(result.serialize(), "[Formyl]-PEPTIDE")
        # Original should be unchanged
        self.assertEqual(annotation.serialize(), "[Acetyl]-PEPTIDE")

    def test_add_nterm_mods_append_to_existing(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        annotation.append_nterm_mod("Formyl", inplace=True)

        self.assertEqual(annotation.serialize(), "[Acetyl][Formyl]-PEPTIDE")

    def test_add_nterm_mods_append_to_empty(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.append_nterm_mod("Acetyl", inplace=True)

        self.assertEqual(annotation.serialize(), "[Acetyl]-PEPTIDE")

    def test_add_nterm_mods_list_input(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.set_nterm_mods(["Acetyl", "Formyl"], inplace=True)

        self.assertEqual(annotation.serialize(), "[Acetyl][Formyl]-PEPTIDE")

    def test_add_nterm_mods_none_input_no_append(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        annotation.set_nterm_mods(None, inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE")

    """
    TESTS FOR: add_cterm_mods
    """

    def test_add_cterm_mods_replace_inplace_true(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE-[Amide]")
        result = annotation.set_cterm_mods("Hydroxyl", inplace=True)

        self.assertIs(result, annotation)
        self.assertEqual(annotation.serialize(), "PEPTIDE-[Hydroxyl]")

    def test_add_cterm_mods_replace_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE-[Amide]")
        result = annotation.set_cterm_mods("Hydroxyl", inplace=False)

        self.assertIsNot(result, annotation)
        self.assertEqual(result.serialize(), "PEPTIDE-[Hydroxyl]")
        # Original should be unchanged
        self.assertEqual(annotation.serialize(), "PEPTIDE-[Amide]")

    def test_add_cterm_mods_append_to_existing(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE-[Amide]")
        annotation.append_cterm_mod("Hydroxyl", inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE-[Amide][Hydroxyl]")

    def test_add_cterm_mods_append_to_empty(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.append_cterm_mod("Amide", inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE-[Amide]")

    def test_add_cterm_mods_list_input(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.set_cterm_mods(["Amide", "Hydroxyl"], inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE-[Amide][Hydroxyl]")

    def test_add_cterm_mods_none_input_no_append(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE-[Amide]")
        annotation.set_cterm_mods(None, inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE")

    """
    TESTS FOR: add_mod_by_type
    """

    def test_add_mod_by_type_string_parameter(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        result = annotation.set_nterm_mods("Acetyl", inplace=True)

        self.assertIs(result, annotation)
        self.assertEqual(annotation.serialize(), "[Acetyl]-PEPTIDE")

    def test_add_mod_by_type_mod_type_parameter(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        result = annotation.set_nterm_mods("Acetyl", inplace=True)

        self.assertIs(result, annotation)
        self.assertEqual(annotation.serialize(), "[Acetyl]-PEPTIDE")

    def test_add_mod_by_type_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        result = annotation.set_nterm_mods("Acetyl", inplace=False)

        self.assertIsNot(result, annotation)
        self.assertEqual(result.serialize(), "[Acetyl]-PEPTIDE")
        self.assertEqual(annotation.serialize(), "PEPTIDE")

    def test_add_mod_by_type_all_types(self):
        # Test each mod type
        test_cases = [
            ("Glycan", "labile", "{Glycan}PEPTIDE"),
            ("Unknown", "unknown", "[Unknown]?PEPTIDE"),
            ("Acetyl", "nterm", "[Acetyl]-PEPTIDE"),
            ("Amide", "cterm", "PEPTIDE-[Amide]"),
            ("15N", "isotope", "<15N>PEPTIDE"),
            ("57@C", "static", "<57@C>PEPTIDE"),
        ]

        for mod_val, mod_type, expected_serialization in test_cases:
            with self.subTest(mod_type=mod_type):
                annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
                # dispatch to the corresponding setter to match the 'labile' pattern
                if mod_type == "labile":
                    annotation.set_labile_mods(mod_val, inplace=True)
                elif mod_type == "unknown":
                    annotation.set_unknown_mods(mod_val, inplace=True)
                elif mod_type == "nterm":
                    annotation.set_nterm_mods(mod_val, inplace=True)
                elif mod_type == "cterm":
                    annotation.set_cterm_mods(mod_val, inplace=True)
                elif mod_type == "isotope":
                    annotation.set_isotope_mods(mod_val, inplace=True)
                elif mod_type == "static":
                    annotation.set_static_mods(mod_val, inplace=True)
                self.assertEqual(annotation.serialize(), expected_serialization)

    def test_add_mod_by_type_charge(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotation.set_charge(2, inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE/2")

    def test_add_mod_by_type_charge_adducts(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE/2")
        annotation.append_charge_adduct("Na+", inplace=True)

        self.assertEqual(annotation.serialize(), "PEPTIDE/2[Na+]")

    def test_add_mod_by_type_internal(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        # append an internal mod (new API appends a single internal mod by default at position 0)
        annotation.append_internal_mod_at_index(0, "Phospho", inplace=True)

        # Internal mods are added at position 0 by default in add_mod_by_type
        self.assertEqual(annotation.serialize(), "P[Phospho]EPTIDE")

    def test_add_mod_by_type_intervals(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        interval = pt.Interval(1, 3, False, [pt.Mod("Phospho", 1)])
        # set interval-style mods using a specific setter
        annotation.set_intervals(interval, inplace=True)

        self.assertEqual(annotation.serialize(), "P(EP)[Phospho]TIDE")

    def test_add_mod_by_type_append_behavior(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        annotation.append_nterm_mod("Formyl", inplace=True)

        self.assertEqual(annotation.serialize(), "[Acetyl][Formyl]-PEPTIDE")

    """
    TESTS FOR: add_mod_dict
    """

    def test_add_mod_dict_inplace_true(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        mod_dict = {"nterm": "Acetyl", "cterm": "Amide", "labile": "Glycan"}
        result = annotation.set_mod_dict(mod_dict, append=False, inplace=True)

        self.assertIs(result, annotation)
        self.assertEqual(annotation.serialize(), "{Glycan}[Acetyl]-PEPTIDE-[Amide]")

    def test_add_mod_dict_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        mod_dict = {"nterm": "Acetyl", "cterm": "Amide"}
        result = annotation.set_mod_dict(mod_dict, append=False, inplace=False)

        self.assertIsNot(result, annotation)
        self.assertEqual(result.serialize(), "[Acetyl]-PEPTIDE-[Amide]")
        self.assertEqual(annotation.serialize(), "PEPTIDE")

    def test_add_mod_dict_with_internal_mods(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        mod_dict = {
            "nterm": "Acetyl",
            1: "Phospho",  # Internal mod at position 1
            3: "Methyl",  # Internal mod at position 3
        }
        annotation.set_mod_dict(mod_dict, append=False, inplace=True)

        self.assertEqual(annotation.serialize(), "[Acetyl]-PE[Phospho]PT[Methyl]IDE")

    def test_add_mod_dict_append_behavior(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        mod_dict = {"nterm": "Formyl", "cterm": "Amide"}
        annotation.set_mod_dict(mod_dict, append=True, inplace=True)

        self.assertEqual(annotation.serialize(), "[Acetyl][Formyl]-PEPTIDE-[Amide]")

    def test_add_mod_dict_replace_behavior(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]")
        mod_dict = {"nterm": "Formyl", "cterm": "Hydroxyl"}
        annotation.set_mod_dict(mod_dict, append=False, inplace=True)

        self.assertEqual(annotation.serialize(), "[Formyl]-PEPTIDE-[Hydroxyl]")

    """
    TESTS FOR: edge cases and validation
    """

    def test_add_methods_preserve_sequence(self):
        original_sequence = "PEPTIDE"

        # Test each add method preserves the sequence
        test_cases = [
            ("set_labile_mods", "Glycan"),
            ("set_unknown_mods", "Unknown"),
            ("set_nterm_mods", "Acetyl"),
            ("set_cterm_mods", "Amide"),
        ]

        for method_name, mod_val in test_cases:
            with self.subTest(method=method_name):
                annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
                method = getattr(annotation, method_name)
                method(mod_val, append=False, inplace=True)
                self.assertEqual(annotation.sequence, original_sequence)

    def test_add_methods_return_self_when_inplace(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")

        methods = [
            ("set_labile_mods", "Glycan"),
            ("set_unknown_mods", "Unknown"),
            ("set_nterm_mods", "Acetyl"),
            ("set_cterm_mods", "Amide"),
        ]

        for method_name, mod_val in methods:
            with self.subTest(method=method_name):
                method = getattr(annotation, method_name)
                result = method(mod_val, append=False, inplace=True)
                self.assertIs(result, annotation)

    def test_add_methods_return_copy_when_not_inplace(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")

        methods = [
            ("set_labile_mods", "Glycan"),
            ("set_unknown_mods", "Unknown"),
            ("set_nterm_mods", "Acetyl"),
            ("set_cterm_mods", "Amide"),
        ]

        for method_name, mod_val in methods:
            with self.subTest(method=method_name):
                method = getattr(annotation, method_name)
                result = method(mod_val, append=False, inplace=False)
                self.assertIsNot(result, annotation)

    def test_add_methods_with_multipliers(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        mod_with_multiplier = pt.Mod("Phospho", 2)

        annotation.set_nterm_mods(mod_with_multiplier, inplace=True)

        self.assertEqual(annotation.serialize(), "[Phospho]^2-PEPTIDE")

    def test_add_empty_list_to_existing_mods(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE")
        original_serialization = annotation.serialize()

        # use extend for appending lists (append single mods, extend lists)
        annotation.extend_nterm_mods([], inplace=True)

        # Adding empty list should not change the serialization
        self.assertEqual(annotation.serialize(), original_serialization)

    def test_complex_mod_dict_scenario(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        complex_mod_dict = {
            "labile": ["Glycan", "HexNAc"],
            "nterm": "Acetyl",
            "cterm": ["Amide", "Hydroxyl"],
            0: "Phospho",  # Internal mod at position 0
            2: ["Methyl", "Oxidation"],  # Multiple internal mods at position 2
            "charge": 2,
            "isotope": "15N",
        }

        annotation.set_mod_dict(complex_mod_dict, append=False, inplace=True)

        # Verify the complex modification produces expected serialization
        self.assertEqual(
            annotation.serialize(),
            "{Glycan}{HexNAc}<15N>[Acetyl]-P[Phospho]EP[Methyl][Oxidation]TIDE-[Amide][Hydroxyl]/2",
        )


if __name__ == "__main__":
    unittest.main()



"""
Tests for parsing modification strings with multiple tags (pipe-separated).
"""

import pytest

import peptacular as pt


class TestParseModificationTags:
    """Tests for ModificationTags.from_string function"""

    def test_returns_tuple(self):
        """Test that ModificationTags.from_string returns a tuple"""
        result = pt.ModificationTags.from_string("Oxidation")
        assert isinstance(result, pt.ModificationTags)
        assert len(result) == 1

    def test_contains_correct_tag(self):
        """Test that tuple contains the correct tag"""
        result = pt.ModificationTags.from_string("UNIMOD:35")
        assert len(result) == 1
        first_tag = result[0]
        assert isinstance(first_tag, pt.TagAccession)
        assert first_tag.accession == "35"


class TestMultipleTags:
    """Tests for parsing modification strings with multiple tags (pipe-separated)"""

    def test_two_tags_name_and_accession(self):
        """Test parsing two tags: name and accession"""
        result = pt.ModificationTags.from_string("Oxidation|UNIMOD:35")
        assert isinstance(result, pt.ModificationTags)
        assert len(result) == 2
        first_tag = result[0]
        second_tag = result[1]
        assert isinstance(first_tag, pt.TagName)
        assert first_tag.name == "Oxidation"
        assert isinstance(second_tag, pt.TagAccession)
        assert second_tag.accession == "35"
        assert second_tag.cv == pt.CV.UNIMOD

    def test_two_tags_accession_and_mass(self):
        """Test parsing accession and mass"""
        result = pt.ModificationTags.from_string("UNIMOD:35|+15.995")
        assert len(result) == 2
        first_tag = result[0]
        second_tag = result[1]
        assert isinstance(first_tag, pt.TagAccession)
        assert first_tag.accession == "35"
        assert isinstance(second_tag, pt.TagMass)
        assert second_tag.mass == pytest.approx(15.995)

    def test_three_tags_name_accession_mass(self):
        """Test parsing three tags: name, accession, and mass"""
        result = pt.ModificationTags.from_string("Oxidation|UNIMOD:35|+15.995")
        assert len(result) == 3
        assert isinstance(result[0], pt.TagName)
        assert result[0].name == "Oxidation"
        assert isinstance(result[1], pt.TagAccession)
        assert result[1].accession == "35"
        assert isinstance(result[2], pt.TagMass)
        assert result[2].mass == pytest.approx(15.995)

    def test_four_tags_complete_oxidation(self):
        """Test parsing all four ways to represent oxidation"""
        result = pt.ModificationTags.from_string("Oxidation|UNIMOD:35|+15.995|Formula:O")
        assert len(result) == 4
        assert isinstance(result[0], pt.TagName)
        assert isinstance(result[1], pt.TagAccession)
        assert isinstance(result[2], pt.TagMass)
        assert isinstance(result[3], pt.ChargedFormula)
        assert result[3].formula[0].element == pt.Element.O

    def test_name_and_info(self):
        """Test parsing name with INFO tag"""
        result = pt.ModificationTags.from_string("Phospho|INFO:probable")
        assert len(result) == 2
        assert isinstance(result[0], pt.TagName)
        assert result[0].name == "Phospho"
        assert isinstance(result[1], pt.TagInfo)
        assert result[1].info == "probable"

    def test_accession_and_info(self):
        """Test parsing accession with INFO tag"""
        result = pt.ModificationTags.from_string("MOD:00046|INFO:confident")
        assert len(result) == 2
        assert isinstance(result[0], pt.TagAccession)
        assert result[0].cv == pt.CV.PSI_MOD
        assert isinstance(result[1], pt.TagInfo)
        assert result[1].info == "confident"

    def test_shorthand_with_mass(self):
        """Test parsing shorthand named mod with mass (Rule 4 + Rule 2)"""
        result = pt.ModificationTags.from_string("U:Oxidation|+15.995")
        assert len(result) == 2
        assert isinstance(result[0], pt.TagName)
        assert result[0].cv == pt.CV.UNIMOD
        assert result[0].name == "Oxidation"
        assert isinstance(result[1], pt.TagMass)
        assert result[1].mass == pytest.approx(15.995)

    def test_formula_with_mass(self):
        """Test parsing formula with mass"""
        result = pt.ModificationTags.from_string("Formula:H-2O-1|+79.966")
        assert len(result) == 2
        assert isinstance(result[0], pt.ChargedFormula)
        assert isinstance(result[1], pt.TagMass)

    def test_whitespace_handling(self):
        """Test that whitespace around pipe is handled correctly"""
        result = pt.ModificationTags.from_string("Oxidation | UNIMOD:35 | +15.995")
        assert len(result) == 3
        assert isinstance(result[0], pt.TagName)
        assert result[0].name == "Oxidation"
        assert isinstance(result[1], pt.TagAccession)
        assert result[1].accession == "35"
        assert isinstance(result[2], pt.TagMass)
        assert result[2].mass == pytest.approx(15.995)

    def test_empty_parts_ignored(self):
        """Test that empty parts from consecutive pipes throw error"""
        with pytest.raises(ValueError):
            _ = pt.ModificationTags.from_string("Oxidation||UNIMOD:35")

    def test_complex_real_world_example(self):
        """Test complex real-world example with multiple tags"""
        result = pt.ModificationTags.from_string("Phosphorylation|MOD:00046|M:Phospho|+79.966|Formula:H P O3|INFO:high confidence")
        assert len(result) == 6
        assert isinstance(result[0], pt.TagName)
        assert result[0].name == "Phosphorylation"
        assert isinstance(result[1], pt.TagAccession)
        assert result[1].cv == pt.CV.PSI_MOD
        assert isinstance(result[2], pt.TagName)
        assert result[2].cv == pt.CV.PSI_MOD  # M: is shorthand for named mods
        assert result[2].name == "Phospho"
        assert isinstance(result[3], pt.TagMass)
        assert isinstance(result[4], pt.ChargedFormula)
        assert isinstance(result[5], pt.TagInfo)

    def test_glycan_with_info(self):
        """Test glycan composition with INFO tag"""
        result = pt.ModificationTags.from_string("Glycan:Hex5HexNAc4|INFO:N-glycan")
        assert len(result) == 2
        # First tag should be GlycanTag
        assert isinstance(result[0], pt.GlycanTag)
        assert len(result[0].components) == 2
        assert all(isinstance(g, pt.GlycanComponent) for g in result[0].components)
        assert isinstance(result[1], pt.TagInfo)

    def test_multiple_accessions_different_cvs(self):
        """Test multiple accessions from different CVs"""
        result = pt.ModificationTags.from_string("UNIMOD:35|MOD:00719|RESID:AA0037")
        assert len(result) == 3
        assert isinstance(result[0], pt.TagAccession)
        assert result[0].cv == pt.CV.UNIMOD
        assert isinstance(result[1], pt.TagAccession)
        assert result[1].cv == pt.CV.PSI_MOD
        assert isinstance(result[2], pt.TagAccession)
        assert result[2].cv == pt.CV.RESID

    def test_mixed_shorthand_and_full(self):
        """Test mixing shorthand named mods and full accessions"""
        result = pt.ModificationTags.from_string("U:Oxidation|MOD:00719|M:Phospho")
        assert len(result) == 3
        assert isinstance(result[0], pt.TagName)
        assert result[0].cv == pt.CV.UNIMOD
        assert isinstance(result[1], pt.TagAccession)
        assert result[1].cv == pt.CV.PSI_MOD
        assert isinstance(result[2], pt.TagName)
        assert result[2].cv == pt.CV.PSI_MOD



# filepath: /workspaces/peptacular/tests/test_proforma_funcs/test_filter.py
import unittest

import peptacular as pt


class TestFilterMods(unittest.TestCase):
    def test_filter_mods_basic_inplace_true(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["nterm"], inplace=True)
        # Should keep only nterm mods
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PEPTIDE")
        self.assertEqual(annotation.serialize(), "[Acetyl]-PEPTIDE")  # Original changed too

    def test_filter_mods_basic_inplace_false(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["nterm"], inplace=False)
        # Should keep only nterm mods
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PEPTIDE")
        self.assertEqual(annotation.serialize(), "[Acetyl]-PEPTIDE-[Amide]")  # Original unchanged

    def test_filter_mods_single_string(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PE[Phospho]PTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods="internal", inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "PE[Phospho]PTIDE")

    def test_filter_mods_single_mod_type(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PE[Phospho]PTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=pt.ModType.INTERNAL, inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "PE[Phospho]PTIDE")

    def test_filter_mods_list_of_strings(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PE[Phospho]PTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["nterm", "cterm"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PEPTIDE-[Amide]")

    def test_filter_mods_list_of_mod_types(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PE[Phospho]PTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=[pt.ModType.NTERM, pt.ModType.CTERM], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PEPTIDE-[Amide]")

    def test_filter_mods_mixed_list(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PE[Phospho]PTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["nterm", pt.ModType.INTERNAL], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PE[Phospho]PTIDE")

    """
    TESTS FOR: filtering different modification types
    """

    def test_filter_mods_keep_labile(self):
        annotation = pt.ProFormaAnnotation.parse("{Glycan}[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["labile"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "{Glycan}PEPTIDE")

    def test_filter_mods_keep_static(self):
        annotation = pt.ProFormaAnnotation.parse("<57@C>[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["static"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "<57@C>PEPTIDE")

    def test_filter_mods_keep_isotope(self):
        annotation = pt.ProFormaAnnotation.parse("<15N>[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["isotope"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "<15N>PEPTIDE")

    def test_filter_mods_keep_unknown(self):
        annotation = pt.ProFormaAnnotation.parse("[Unknown]?[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["unknown"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "[Unknown]?PEPTIDE")

    def test_filter_mods_keep_charge(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]/2")
        filtered_annotation = annotation.filter_mods(mods=["charge"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "PEPTIDE/2")

    def test_filter_mods_keep_charge_adducts(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]/[Na:z+1]")
        filtered_annotation = annotation.filter_mods(mods=["charge"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "PEPTIDE/[Na:z+1]")

    def test_filter_mods_keep_intervals(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-P(EP)[Phospho]TIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["interval"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "P(EP)[Phospho]TIDE")

    """
    TESTS FOR: filtering multiple modification types
    """

    def test_filter_mods_keep_multiple_types(self):
        annotation = pt.ProFormaAnnotation.parse("<[57]@C>{Glycan}[Acetyl]-PE[Phospho]PTIDE-[Amide]/2")
        filtered_annotation = annotation.filter_mods(mods=["nterm", "internal", "charge"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PE[Phospho]PTIDE/2")

    def test_filter_mods_keep_all_terminal_mods(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PE[Phospho]PTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["nterm", "cterm"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PEPTIDE-[Amide]")

    """
    TESTS FOR: edge cases
    """

    def test_filter_mods_no_mods_to_filter(self):
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        filtered_annotation = annotation.filter_mods(mods=["nterm"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "PEPTIDE")

    def test_filter_mods_none_parameter(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=None, inplace=False)
        # Should remove all mods when mods=None (empty list)
        self.assertEqual(filtered_annotation.serialize(), "[Acetyl]-PEPTIDE-[Amide]")

    def test_filter_mods_empty_list(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-PEPTIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=[], inplace=False)
        # Should remove all mods when mods=[] (empty list)
        self.assertEqual(filtered_annotation.serialize(), "PEPTIDE")

    def test_filter_mods_all_mod_types(self):
        # Test keeping all modification types (should not change anything)
        annotation = pt.ProFormaAnnotation.parse("<57@C>{Glycan}[Unknown]?[Acetyl]-PE[Phospho]PTIDE-[Amide]/[Na:z+1]")
        all_mod_types = [
            "labile",
            "static",
            "unknown",
            "nterm",
            "cterm",
            "internal",
            "charge",
        ]
        filtered_annotation = annotation.filter_mods(mods=all_mod_types, inplace=False)
        self.assertEqual(filtered_annotation.serialize(), annotation.serialize())

    def test_filter_mods_complex_annotation(self):
        annotation = pt.ProFormaAnnotation.parse("<[57]@C>{Glycan}[Acetyl]-P(EP)[Phospho]TI[Methyl]DE-[Amide]/2")
        # Keep only internal modifications
        filtered_annotation = annotation.filter_mods(mods=["internal"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "PEPTI[Methyl]DE")

    def test_filter_mods_with_intervals_ambiguous(self):
        annotation = pt.ProFormaAnnotation.parse("[Acetyl]-P(?EP)TIDE-[Amide]")
        filtered_annotation = annotation.filter_mods(mods=["interval"], inplace=False)
        self.assertEqual(filtered_annotation.serialize(), "P(?EP)TIDE")


if __name__ == "__main__":
    unittest.main()



"""
Parallel Execution Examples
============================
Examples of using parallel processing with peptacular sequence functions.
Most functions in the sequence module accept lists and automatically use
multiprocessing for better performance on large datasets.
"""

import peptacular as pt
import time

# ============================================================================
# BASIC PARALLEL EXECUTION (must be run through __main__)
# ============================================================================

def run():
    print("=" * 60)
    print("BASIC PARALLEL EXECUTION")
    print("=" * 60)

    # Create a list of peptide sequences
    peptides = [
        'PEPTIDE',
        'PATRICK',
        'TYLER',
        'GARRETT'
    ]

    # Parse multiple sequences in parallel
    # Pass a list of strings - automatically uses multiprocessing
    annotations = pt.parse(peptides)
    print(f"\nParsed {len(annotations)} peptides:")
    for annot in annotations:
        print(f"  {annot.serialize()}")

    # ============================================================================
    # MASS CALCULATIONS IN PARALLEL
    # ============================================================================

    print("\n" + "=" * 60)
    print("MASS CALCULATIONS")
    print("=" * 60)

    # Calculate masses for multiple peptides
    masses = pt.mass(peptides)
    print("\nMasses:")
    for pep, mass in zip(peptides, masses):
        print(f"  {pep}: {mass:.4f} Da")

    # With charge states
    masses_charged = pt.mass(peptides, charge=2)
    print("\nMasses at +2 charge:")
    for pep, mass in zip(peptides, masses_charged):
        print(f"  {pep}: {mass:.4f} Da")

    # m/z calculations
    mz_values = pt.mz(peptides, charge=2)
    print("\nm/z at +2 charge:")
    for pep, mz in zip(peptides, mz_values):
        print(f"  {pep}: {mz:.4f}")

    # ============================================================================
    # COMPOSITION IN PARALLEL
    # ============================================================================

    print("\n" + "=" * 60)
    print("COMPOSITION CALCULATIONS")
    print("=" * 60)

    # Get compositions for multiple peptides
    compositions = pt.comp(peptides)
    print("\nCompositions:")
    for pep, comp in zip(peptides, compositions):
        comp_str = {str(elem): count for elem, count in comp.items()}
        print(f"  {pep}: {comp_str}")

    # ============================================================================
    # PHYSICOCHEMICAL PROPERTIES
    # ============================================================================

    print("\n" + "=" * 60)
    print("PHYSICOCHEMICAL PROPERTIES")
    print("=" * 60)

    # Calculate hydrophobicity for all peptides
    hydrophobicity = pt.hydrophobicity(peptides)
    print("\nHydrophobicity:")
    for pep, hydro in zip(peptides, hydrophobicity):
        print(f"  {pep}: {hydro:.3f}")

    # Calculate pI values
    pi_values = pt.pi(peptides)
    print("\nIsoelectric points:")
    for pep, pi in zip(peptides, pi_values):
        print(f"  {pep}: {pi:.2f}")

    # Calculate aromaticity
    aromaticity = pt.aromaticity(peptides)
    print("\nAromaticity:")
    for pep, arom in zip(peptides, aromaticity):
        print(f"  {pep}: {arom:.3f}")

    # ============================================================================
    # USING ANNOTATION OBJECTS
    # ============================================================================

    print("\n" + "=" * 60)
    print("USING ANNOTATION OBJECTS")
    print("=" * 60)

    # Can also pass lists of ProFormaAnnotation objects
    modified_peptides = [
        pt.parse('[Acetyl]-PEPTIDE'),
        pt.parse('PEM[Oxidation]TIDE'),
        pt.parse('SEQS[Phospho]UENCE/2')
    ]

    # Calculate masses from annotations
    annot_masses = pt.mass(modified_peptides)
    print("\nModified peptide masses:")
    for annot, mass in zip(modified_peptides, annot_masses):
        print(f"  {annot.serialize()}: {mass:.4f} Da")

    # ============================================================================
    # CUSTOM PROPERTY CALCULATIONS
    # ============================================================================

    print("\n" + "=" * 60)
    print("CUSTOM PROPERTY CALCULATIONS")
    print("=" * 60)

    # Calculate custom properties in parallel
    custom_props = pt.calc_property(
        peptides,
        scale=pt.HydrophobicityScale.KYTE_DOOLITTLE,
        aggregation_method='avg'
    )
    print("\nKyte-Doolittle hydrophobicity (average):")
    for pep, prop in zip(peptides, custom_props):
        print(f"  {pep}: {prop:.3f}")

    # ============================================================================
    # PERFORMANCE COMPARISON
    # ============================================================================

    print("\n" + "=" * 60)
    print("PERFORMANCE COMPARISON")
    print("=" * 60)

    # Create larger dataset for timing
    large_dataset = peptides * 10000  # 40,000 peptides

    # Time parallel execution
    start = time.time()
    _ = pt.mass(large_dataset)
    parallel_time = time.time() - start

    print(f"\nProcessed {len(large_dataset)} peptides:")
    print(f"  Parallel execution: {parallel_time:.3f} seconds")
    print(f"  Average per peptide: {parallel_time/len(large_dataset)*1000:.2f} ms")

    # verse time serial execution
    start = time.time()
    _ = [pt.mass(pep) for pep in large_dataset]
    serial_time = time.time() - start   
    print(f"  Serial execution: {serial_time:.3f} seconds")
    print(f"  Average per peptide: {serial_time/len(large_dataset)*1000:.2f} ms")

    print("\n" + "=" * 60)

if __name__ == "__main__":
    run()


"""
Tests for parsing custom modification tags.
"""

import pytest

import peptacular as pt


class TestTagCustom:
    """Tests for pt.TagCustom.from_string() function"""

    def test_simple_custom_tag(self):
        """Test parsing simple custom tag"""
        result = pt.TagCustom.from_string("C:MyCustomMod")
        assert isinstance(result, pt.TagCustom)
        assert result.name == "MyCustomMod"

    def test_custom_tag_with_underscores(self):
        """Test custom tag with underscores"""
        result = pt.TagCustom.from_string("C:my_custom_mod")
        assert result.name == "my_custom_mod"

    def test_custom_tag_with_numbers(self):
        """Test custom tag with numbers"""
        result = pt.TagCustom.from_string("C:CustomMod123")
        assert result.name == "CustomMod123"

    def test_custom_tag_with_special_chars(self):
        """Test custom tag with special characters"""
        result = pt.TagCustom.from_string("C:Mod-v2.0")
        assert result.name == "Mod-v2.0"

    def test_missing_prefix_raises_error(self):
        """Test that missing C: prefix raises ValueError"""
        with pytest.raises(ValueError):
            pt.TagCustom.from_string("CustomMod")

    def test_empty_name_after_prefix(self):
        """Test custom tag with empty name after prefix"""
        # This should work - name is empty string
        result = pt.TagCustom.from_string("C:")
        assert result.name == ""

    def test_string_representation(self):
        """Test string representation of custom tag"""
        result = pt.TagCustom.from_string("C:MyMod")
        assert str(result) == "C:MyMod"

    def test_via_modification_string(self):
        """Test parsing custom tag via parse_modification_string"""
        result = pt.ModificationTags.from_string("C:MyCustomMod").tags[0]
        assert isinstance(result, pt.TagCustom)
        assert result.name == "MyCustomMod"



import unittest
from types import GeneratorType

import peptacular as pt


class TestSlidingWindow(unittest.TestCase):
    def test_basic_sliding_window(self):
        # Test slicing a ProFormaAnnotation
        annotation = pt.ProFormaAnnotation(sequence="PEPTIDE")

        sliding_windows = annotation.sliding_windows(5)

        # ensure generator
        self.assertIsInstance(sliding_windows, GeneratorType)

        for i, window in enumerate(sliding_windows):
            if i == 0:
                self.assertEqual(window.sequence, "PEPTI")
            elif i == 1:
                self.assertEqual(window.sequence, "EPTID")
            elif i == 2:
                self.assertEqual(window.sequence, "PTIDE")
            else:
                break

    def test_reverse_sliding_window(self):
        # Test reverse sliding windows
        annotation = pt.ProFormaAnnotation(sequence="PEPTIDE")

        sliding_windows = annotation.sliding_windows(5, reverse=True)

        # ensure generator
        self.assertIsInstance(sliding_windows, GeneratorType)

        for i, window in enumerate(sliding_windows):
            if i == 0:
                self.assertEqual(window.sequence, "PTIDE")
            elif i == 1:
                self.assertEqual(window.sequence, "EPTID")
            elif i == 2:
                self.assertEqual(window.sequence, "PEPTI")
            else:
                break



"""Tests for pt.SequenceElement and pt.SequenceRegion"""

import unittest

import peptacular as pt


class TestSequenceElement(unittest.TestCase):
    """Test cases for pt.SequenceElement"""

    def test_simple_amino_acid(self):
        """Test sequence element with just an amino acid"""
        se = pt.SequenceElement(amino_acid=pt.AminoAcid.M)
        self.assertEqual(se.amino_acid, pt.AminoAcid.M)
        self.assertEqual(len(se.modifications), 0)
        self.assertEqual(str(se), "M")

    def test_amino_acid_with_single_modification(self):
        """Test sequence element with one modification"""
        mod = pt.ModificationTags.from_string("Oxidation")
        se = pt.SequenceElement(amino_acid=pt.AminoAcid.M, modifications=(mod,))
        self.assertEqual(se.amino_acid, pt.AminoAcid.M)
        self.assertEqual(len(se.modifications), 1)
        self.assertEqual(str(se), "M[Oxidation]")


class TestSequenceRegion(unittest.TestCase):
    """Test cases for pt.SequenceRegion"""

    def test_simple_sequence_region(self):
        """Test sequence region with no modifications"""
        se1 = pt.SequenceElement(amino_acid=pt.AminoAcid.P)
        se2 = pt.SequenceElement(amino_acid=pt.AminoAcid.E)
        se3 = pt.SequenceElement(amino_acid=pt.AminoAcid.P)

        sr = pt.SequenceRegion(sequence=(se1, se2, se3), modifications=(), ambiguous=False)

        self.assertEqual(len(sr.sequence), 3)
        self.assertEqual(len(sr.modifications), 0)
        self.assertEqual(str(sr), "(PEP)")

    def test_sequence_region_with_modification(self):
        """Test sequence region with one modification"""
        se1 = pt.SequenceElement(amino_acid=pt.AminoAcid.P)
        se2 = pt.SequenceElement(amino_acid=pt.AminoAcid.E)
        se3 = pt.SequenceElement(amino_acid=pt.AminoAcid.P)

        mod = pt.ModificationTags.from_string("Phospho")

        sr = pt.SequenceRegion(sequence=(se1, se2, se3), modifications=(mod,), ambiguous=False)

        self.assertEqual(len(sr.sequence), 3)
        self.assertEqual(len(sr.modifications), 1)
        self.assertEqual(str(sr), "(PEP)[Phospho]")



from tacular import FRAGMENT_ION_LOOKUP, FragmentIonInfo, IonType, IonTypeLiteral


def _validate_position(ion_type: IonType, position: int | tuple[int, int] | None, sequence_length: int) -> tuple[int, int] | None:
    """Validate fragment position based on ion type."""

    ion_info: FragmentIonInfo = FRAGMENT_ION_LOOKUP[ion_type]

    if position is None:
        return None  # whole sequence, no position-specific adjustments

    if isinstance(position, int):
        if ion_info.is_forward:
            return (0, position)
        if ion_info.is_backward:
            return (sequence_length - position, sequence_length)
        if ion_info.ion_type == IonType.IMMONIUM:
            return (position - 1, position)
        raise ValueError(f"Integer position {position} not valid for ion type {ion_type.name}, which is not a forward or backward ion.")

    if isinstance(position, tuple):
        if ion_info.is_internal:
            return (position[0] - 1, position[1])
        raise ValueError(f"Tuple position {position} not valid for ion type {ion_type.name}, which is not an internal ion.")

    raise ValueError(f"Invalid position type: {type(position)} for ion type {ion_type.name}.")


def validate_position(ion_type: IonType | IonTypeLiteral, position: int | tuple[int, int] | None, sequence_length: int) -> tuple[int, int] | None:
    """
    Validates the passed position and return the true zero index position tuple for the fragment.
    """
    pos: tuple[int, int] | None = _validate_position(IonType(ion_type), position, sequence_length)
    if pos is None:
        return None
    start, end = pos
    if start < 0 or end > sequence_length or start >= end:
        raise ValueError(f"Invalid internal fragment position: {pos} for sequence length {sequence_length}.")
    return pos



import random
import time

import peptacular as pt


def generate_random_protein(length: int = 500) -> str:
    """Generate a random protein sequence."""
    amino_acids = "ACDEFGHIKLMNPQRSTVWY"
    return "".join(random.choice(amino_acids) for _ in range(length))


def benchmark_digest_methods():
    """Benchmark all three parallelization methods."""

    # Generate test data
    proteins = [generate_random_protein(1_000) for _ in range(1000)]

    methods = ["sequential", "thread", "process"]
    results = {}

    print("=" * 80)
    print("BENCHMARKING DIGEST METHODS")
    print("=" * 80)
    print(f"Test data: {len(proteins)} proteins, ~1000 AA each")
    print("Parameters: cleave_on='KR', missed_cleavages=2, min_len=6, max_len=50")
    print("=" * 80)

    for method in methods:
        print(f"\nTesting method: {method.upper()}")
        print("-" * 40)

        # Warm-up run (not timed)
        _ = list(
            pt.digest(
                proteins[:10],
                cleave_on="KR",
                missed_cleavages=2,
                semi=False,
                min_len=6,
                max_len=50,
                method=method,
            )
        )

        # Actual benchmark
        start_time = time.time()
        peptides = list(
            pt.digest(
                proteins,
                cleave_on="KR",
                missed_cleavages=2,
                semi=False,
                min_len=6,
                max_len=50,
                method=method,
            )
        )
        elapsed = time.time() - start_time

        total_peptides = len(peptides)
        results[method] = {
            "time": elapsed,
            "peptides": total_peptides,
            "ms_per_protein": elapsed / len(proteins) * 1000,
        }

        print(f"  Time: {elapsed:.3f} seconds")
        print(f"  Peptides generated: {total_peptides:,}")
        print(f"  Speed: {elapsed / len(proteins) * 1000:.2f} ms per protein")
        print(f"  Throughput: {len(proteins) / elapsed:.1f} proteins/sec")

    # Summary comparison
    print("\n" + "=" * 80)
    print("SUMMARY")
    print("=" * 80)

    fastest_method = min(results, key=lambda x: results[x]["time"])
    fastest_time = results[fastest_method]["time"]

    print(f"{'Method':<12} {'Time (s)':<12} {'Speedup':<12} {'ms/protein':<15}")
    print("-" * 80)

    for method in methods:
        time_val = results[method]["time"]
        speedup = results["sequential"]["time"] / time_val if method != "sequential" else 1.0
        ms_per_protein = results[method]["ms_per_protein"]

        marker = " ← FASTEST" if method == fastest_method else ""
        print(f"{method:<12} {time_val:<12.3f} {speedup:<12.2f}x {ms_per_protein:<15.2f}{marker}")

    print("=" * 80)

    return results


if __name__ == "__main__":
    benchmark_digest_methods()



from collections import Counter
from collections.abc import Mapping
from typing import Any, Literal

from tacular import (
    ELEMENT_LOOKUP,
    FRAGMENT_ION_LOOKUP,
    ElementInfo,
    IonType,
    IonTypeProperty,
)

from ..constants import ModType
from ..proforma_components import (
    ChargedFormula,
    GlobalChargeCarrier,
)
from .mod import Mods
from .positions import validate_position

# Maps internal ion type value tuples to their neutral loss diff relative to "by" (the default internal fragment).
# None means no difference from "by".
_INTERNAL_MASS_DIFFS: dict[tuple[str, str], str | None] = {
    ("a", "x"): None,
    ("b", "x"): "+CO",
    ("c", "x"): "+CHNO",
    ("a", "y"): "-CO",
    ("b", "y"): None,
    ("c", "y"): "+NH",
    ("a", "z"): "-CHNO",
    ("b", "z"): "-NH",
    ("c", "z"): None,
}

# Maps peptacular variant ion types to their canonical mzPAF series string.
_ION_TYPE_TO_MZPAF_SERIES: dict[IonType, str] = {
    IonType.W_VALINE: "w",
    IonType.D_VALINE: "d",
    IonType.WB_ISOLEUCINE: "wb",
    IonType.WB_THREONINE: "wb",
    IonType.WA_ISOLEUCINE: "wa",
    IonType.WA_THREONINE: "wa",
    IonType.DB_ISOLEUCINE: "db",
    IonType.DB_THREONINE: "db",
    IonType.DA_ISOLEUCINE: "da",
    IonType.DA_THREONINE: "da",
}


class Fragment:
    def __init__(
        self,
        ion_type: IonType,
        position: int | tuple[int, int] | None,
        mass: float,
        monoisotopic: bool,
        charge_state: int,
        charge_adducts: tuple[str, ...] | None = None,
        isotopes: Mapping[str, int] | int | None = None,
        deltas: Mapping[str | float, int] | None = None,
        composition: Mapping[ElementInfo, int] | None = None,
        parent_sequence: str | None = None,
        parent_sequence_length: int | None = None,
    ) -> None:
        self.ion_type: IonType = ion_type
        self.position: int | tuple[int, int] | None = position
        self.mass: int | float = mass
        self.monoisotopic: bool = monoisotopic
        self.charge_state: int = charge_state
        # If None and charge_state != 0: means protonated
        self._charge_adducts: tuple[str, ...] | None = charge_adducts
        # int means 13C count
        self._isotopes: Mapping[str, int] | int | None = isotopes
        self._losses: Mapping[str | float, int] | None = deltas
        # Optional composition cache
        self._composition: Counter[ElementInfo] | None = composition
        self.parent_sequence: str | None = parent_sequence
        self.parent_sequence_length: int | None = parent_sequence_length

    @property
    def composition(self) -> Counter[ElementInfo] | None:
        if self._composition is not None:
            return self._composition

        if self.parent_sequence is None:
            raise ValueError("Cannot calculate composition without parent sequence or explicit composition")

        if self.parent_sequence_length is None:
            raise ValueError("Cannot calculate composition without parent sequence length")

        from .annotation import ProFormaAnnotation

        annot = ProFormaAnnotation.parse(self.parent_sequence)

        pos = validate_position(self.ion_type, self.position, self.parent_sequence_length)
        if pos is not None:
            start, end = pos
            annot = annot[slice(start, end)]

        return annot.comp(isotopes=self.isotopes, deltas=self.losses, charge=self.charge_state if self._charge_adducts is None else self.charge_adducts)  # type: ignore

    @property
    def mz(self) -> float:
        return self.mass / abs(self.charge_state) if self.charge_state != 0 else self.mass

    @property
    def neutral_mass(self) -> float:
        # subract adduct masses
        total_adduct_mass = 0.0
        for adduct in self.charge_adducts:
            total_adduct_mass += adduct.get_mass(self.monoisotopic)
        return self.mass - total_adduct_mass

    @property
    def charge_adducts(self) -> Mods[GlobalChargeCarrier]:
        if self._charge_adducts is None:
            if self.charge_state != 0:
                return Mods[GlobalChargeCarrier](
                    mod_type=ModType.CHARGE,
                    _mods={GlobalChargeCarrier.charged_proton(self.charge_state).serialize(): 1},
                )
            # no adducts no charge
            return Mods[GlobalChargeCarrier](mod_type=ModType.CHARGE, _mods={})

        # we have adducts, convert to Mods object
        else:
            return Mods[GlobalChargeCarrier](mod_type=ModType.CHARGE, _mods={k: 1 for k in self._charge_adducts})

    @property
    def is_protonated(self) -> bool:
        if self._charge_adducts is None and self.charge_state != 0:
            return True
        return False

    @property
    def isotopes(self) -> Mapping[ElementInfo, int]:
        if self._isotopes is not None:
            if isinstance(self._isotopes, int):
                return {ELEMENT_LOOKUP["13C"]: self._isotopes}

            if isinstance(self._isotopes, dict):
                return {ELEMENT_LOOKUP[elem]: count for elem, count in self._isotopes.items()}

        return {}

    @property
    def is_c13(self) -> bool:
        if self._isotopes is not None and isinstance(self._isotopes, int):
            return True
        return False

    @property
    def losses(self) -> Mapping[ChargedFormula | float, int]:
        if self._losses is not None:
            losses = {}
            for loss_name, count in self._losses.items():
                if isinstance(loss_name, float | int):
                    losses[loss_name] = count
                    continue
                loss_formula = ChargedFormula.from_string(loss_name, require_formula_prefix=False)
                losses[loss_formula] = count
            return losses
        return {}

    def asdict(self) -> dict[str, Any]:
        return {
            "ion_type": self.ion_type,
            "position": self.position,
            "mass": self.mass,
            "charge_state": self.charge_state,
            "monoisotopic": self.monoisotopic,
            "charge_adducts": self.charge_adducts,
            "isotopes": self.isotopes,
            "losses": self.losses,
        }

    def serialize(self, format: Literal["default", "mzpaf"] = "default", include_sequence: bool = True) -> str:
        """Serialize the fragment to a string representation.

        :param format: Output format. ``"default"`` returns the human-readable representation,
            ``"mzpaf"`` returns the mzPAF (Peak Annotation Format) label string.
        :type format: Literal["default", "mzpaf"]
        :param include_sequence: If True, include the peptide sequence in the mzPAF label. Only used for ``"mzpaf"`` format.
        :type include_sequence: bool
        :return: The serialized fragment string.
        :rtype: str
        """
        if format == "default":
            return str(self)
        elif format == "mzpaf":
            return self._serialize_mzpaf(include_sequence=include_sequence)
        else:
            raise ValueError(f"Unknown format: {format!r}. Use 'default' or 'mzpaf'.")

    def to_mzpaf(self, include_sequence: bool = True) -> str:
        """Serialize the fragment to an mzPAF (Peak Annotation Format) label string.

        :param include_sequence: If True, include the peptide sequence in the label.
        :type include_sequence: bool
        :return: The mzPAF label string (e.g. ``"y3{IDE}^2"``).
        :rtype: str
        """
        return self._serialize_mzpaf(include_sequence=include_sequence)

    def _serialize_mzpaf(self, include_sequence: bool = True) -> str:
        """Build the mzPAF label string for this fragment."""
        from .annotation import ProFormaAnnotation

        parts: list[str] = []
        internal_loss: str | None = None

        if self.ion_type is None:
            parts.append("?")
        else:
            ion_info = FRAGMENT_ION_LOOKUP[self.ion_type]

            if ion_info.properties & (IonTypeProperty.FORWARD | IonTypeProperty.BACKWARD):
                # Peptide series ions (a, b, c, x, y, z, d, w, da, db, wa, wb)
                series_str = _ION_TYPE_TO_MZPAF_SERIES.get(ion_info.ion_type, ion_info.ion_type.value)
                position = self.position if isinstance(self.position, int) else -1
                parts.append(f"{series_str}{position}")

                if include_sequence and self.parent_sequence is not None:
                    seq = self.sequence
                    if seq is not None:
                        seq_no_charge = ProFormaAnnotation.parse(seq).serialize(exclude_charge=True)
                        parts.append(f"{{{seq_no_charge}}}")

            elif ion_info.properties & IonTypeProperty.INTERNAL:
                if ion_info.id == IonType.IMMONIUM:
                    # Immonium ion: I{amino_acid}[{modification}]
                    if self.parent_sequence is not None:
                        seq = self.sequence
                        if seq is not None:
                            annot = ProFormaAnnotation.parse(seq)
                            parts.append(f"I{annot.sequence}")

                            if annot.has_internal_mods_at_index(0):
                                internal_mods = annot.get_internal_mods_at_index(0)
                                if len(internal_mods) > 1:
                                    raise ValueError(f"Multiple internal mods on immonium ion not supported in mzPAF, got {internal_mods}")
                                if len(internal_mods) == 1 and internal_mods.mods[0].count > 1:
                                    raise ValueError(f"Multiple occurrences of internal mod on immonium ion not supported in mzPAF, got {internal_mods}")
                                mods_str = internal_mods.serialize()[1:-1]  # remove surrounding brackets
                                if mods_str == "":
                                    raise ValueError(f"Empty modification string for immonium ion is not valid in mzPAF. Internal mods: {internal_mods}")
                                parts.append(f"[{mods_str}]")
                        else:
                            raise ValueError("Immonium ion must have a sequence annotation.")
                    else:
                        raise ValueError("Immonium ion must have a parent sequence.")
                else:
                    # Internal fragment: m{start}:{end}[{sequence}]
                    if isinstance(self.position, tuple) and len(self.position) == 2:
                        start, end = self.position
                    else:
                        start, end = -1, -1

                    parts.append(f"m{start}:{end}")

                    if include_sequence and self.parent_sequence is not None:
                        seq = self.sequence
                        if seq is not None:
                            seq_no_charge = ProFormaAnnotation.parse(seq).serialize(exclude_charge=True)
                            parts.append(f"{{{seq_no_charge}}}")

                    # Add internal mass diff neutral loss
                    internal_ion_key = tuple(list(ion_info.ion_type.value))
                    if internal_ion_key in _INTERNAL_MASS_DIFFS:
                        internal_loss = _INTERNAL_MASS_DIFFS[internal_ion_key]
                    else:
                        raise ValueError(f"Internal ion type {ion_info.ion_type} not supported in mzPAF.")

            elif ion_info.properties & IonTypeProperty.INTACT:
                if ion_info.ion_type == IonType.PRECURSOR:
                    parts.append("p")
                else:
                    raise ValueError(f"Cannot convert intact ion type {ion_info.id} to mzPAF.")
            else:
                raise ValueError(f"Cannot convert fragment with ion type {self.ion_type} to mzPAF.")

        # Neutral losses from self.losses
        if self._losses is not None:
            for loss_key, count in self._losses.items():
                if isinstance(loss_key, float | int):
                    mass_val = float(loss_key) * count
                    parts.append(f"{mass_val:+.5f}")
                else:
                    loss_formula = ChargedFormula.from_string(loss_key, require_formula_prefix=False)
                    paf_formula = loss_formula.to_mz_paf()
                    sign = paf_formula[0]
                    if sign not in ("+", "-"):
                        raise ValueError(f"Invalid formula sign in loss: {paf_formula}")
                    mult = 1 if sign == "+" else -1
                    abs_count = abs(count * mult)
                    count_str = str(abs_count) if abs_count > 1 else ""
                    parts.append(f"{sign}{count_str}{paf_formula[1:]}")

        # Internal mass diff loss (from internal fragment type)
        if internal_loss is not None:
            parts.append(internal_loss)

        # Isotopes
        if self._isotopes is not None:
            if isinstance(self._isotopes, int):
                count_str = str(self._isotopes) if self._isotopes > 1 else ""
                parts.append(f"+{count_str}i")
            elif isinstance(self._isotopes, dict):
                for elem, count in self._isotopes.items():
                    count_str = str(count) if count > 1 else ""
                    parts.append(f"+{count_str}i{elem}")

        # Adducts
        if self._charge_adducts is not None:
            adduct_parts: list[str] = []
            for mod in self.charge_adducts.mods:
                carrier: GlobalChargeCarrier = mod.value
                # to_mz_paf() returns "M+Na", we strip the "M" prefix
                paf_str = carrier.to_mz_paf()
                adduct_parts.append(paf_str[1:])  # strip "M", keep "+Na"
            parts.append(f"[M{''.join(adduct_parts)}]")

        # Charge: omit for +1 (implicit); include for everything else
        if self.charge_state != 0 and self.charge_state != 1:
            parts.append(f"^{self.charge_state}")

        return "".join(parts)

    def __str__(self) -> str:
        parts = []
        parts.append(f"ion_type={self.ion_type}")

        if self.position is not None:
            parts.append(f"position={self.position}")

        parts.append(f"mass={self.mass:.4f}")
        parts.append(f"charge={self.charge_state}")

        # Only show charge_adducts if not simple protonation
        if not self.is_protonated and self.charge_state != 0:
            parts.append(f"charge_adducts={self.charge_adducts}")

        if self.isotopes:
            parts.append(f"isotopes={dict(self.isotopes)}")

        if self.losses:
            parts.append(f"losses={dict(self.losses)}")

        return f"Fragment({', '.join(parts)})"

    def __repr__(self) -> str:
        return (
            f"Fragment(ion_type={self.ion_type!r}, position={self.position!r}, "
            f"mass={self.mass}, monoisotopic={self.monoisotopic}, "
            f"charge_state={self.charge_state}, charge_adducts={self._charge_adducts!r}, "
            f"isotopes={self._isotopes!r}, deltas={self._losses!r}, "
            f"composition={self._composition!r}, parent_sequence={self.parent_sequence!r}, "
            f"parent_sequence_length={self.parent_sequence_length})"
        )

    @property
    def sequence(self) -> str | None:
        if self.parent_sequence is None:
            raise ValueError("Cannot determine fragment sequence without parent sequence")

        if self.parent_sequence_length is None:
            raise ValueError("Cannot determine fragment sequence without parent sequence length")

        pos = validate_position(self.ion_type, self.position, self.parent_sequence_length)
        if pos is None:
            return self.parent_sequence
        elif isinstance(pos, tuple):
            from .annotation import ProFormaAnnotation

            start, end = pos
            return (
                ProFormaAnnotation.parse(self.parent_sequence)[slice(start, end)]
                .set_charge(self.charge_state if self._charge_adducts is None else self.charge_adducts)
                .serialize()
            )

        raise ValueError("Invalid position format for fragment sequence extraction")



from __future__ import annotations

import re
from collections import Counter
from collections.abc import Iterable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .annotation import ProFormaAnnotation


def condense_mods_to_intervals(annotation: ProFormaAnnotation, inplace: bool = True) -> ProFormaAnnotation:
    """
    Condense modifications into intervals
    """
    if inplace is False:
        return condense_mods_to_intervals(annotation.copy(), inplace=True)

    if not annotation.has_intervals or not annotation.has_internal_mods:
        return annotation

    indexes_to_remove: list[int] = []
    for index, mods in annotation.internal_mods.items():
        for interval in annotation.intervals:
            if interval.start <= index < interval.end:
                # Merge modifications from internal_mods into interval
                for mod in mods:
                    interval.append_mod(mod, inplace=True)

                indexes_to_remove.append(index)

    for index in indexes_to_remove:
        annotation.clear_internal_mod_at_index(index, inplace=True)

    return annotation


def coverage(
    annotation: ProFormaAnnotation,
    annotations: Iterable[ProFormaAnnotation],
    accumulate: bool = False,
    ignore_mods: bool = False,
    ignore_ambiguity: bool = False,
) -> list[int]:
    cov_arr = [0] * len(annotation)

    for sub_annots in annotations:
        peptide_cov = [1] * len(sub_annots)
        if ignore_ambiguity is False:
            # Filter for ambiguous intervals only
            for interval in sub_annots.intervals:
                if interval.ambiguous:
                    for i in range(interval.start, interval.end):
                        peptide_cov[i] = 0

        for subsequence_index in find_indices(sub_annots, other=annotation, ignore_mods=ignore_mods):
            start = subsequence_index

            for i, cov in enumerate(peptide_cov):
                if accumulate:
                    cov_arr[start + i] += cov
                else:
                    cov_arr[start + i] = cov

    # apply to annotation
    if ignore_ambiguity is False:
        # Set coverage to 0 for ambiguous positions
        for interval in annotation.intervals:
            if interval.ambiguous:
                for i in range(interval.start, interval.end):
                    cov_arr[i] = 0

    return cov_arr


def percent_coverage(
    annotation: ProFormaAnnotation,
    annotations: Iterable[ProFormaAnnotation],
    accumulate: bool = False,
    ignore_mods: bool = False,
    ignore_ambiguity: bool = False,
) -> float:
    cov_arr = coverage(
        annotation,
        annotations,
        accumulate=accumulate,
        ignore_mods=ignore_mods,
        ignore_ambiguity=ignore_ambiguity,
    )
    if len(cov_arr) == 0:
        return 0.0

    return sum(cov_arr) / len(cov_arr)


def modification_coverage(
    annotation: ProFormaAnnotation,
    annotations: Iterable[ProFormaAnnotation],
    ignore_ambiguity: bool = False,
    accumulate: bool = False,
) -> dict[int, int]:
    cov_dict = coverage(
        annotation,
        annotations,
        accumulate=accumulate,
        ignore_mods=False,
        ignore_ambiguity=ignore_ambiguity,
    )
    mod_cov: dict[int, int] = {}
    for pos in annotation.internal_mods.keys():
        mod_cov[pos] = cov_dict[pos]

    return mod_cov


def condense_static_mods(annotation: ProFormaAnnotation, inplace: bool = True) -> ProFormaAnnotation:
    """
    Condense static mods into internal mods

    Note: This function currently has limited functionality as the static mod
    parsing is not yet fully implemented in the new structure.
    """
    if inplace is False:
        return condense_static_mods(annotation.copy(), inplace=True)

    if not annotation.has_static_mods:
        return annotation

    static_mod_map = annotation.map_static_mods_to_indexes()
    for index, mods in static_mod_map.items():
        if index == -2:  # c term
            annotation.extend_cterm_mods(mods, inplace=True)
        elif index == -1:  # n term
            annotation.extend_nterm_mods(mods, inplace=True)
        else:
            annotation.extend_internal_mods_at_index(index, mods, inplace=True)

    annotation.clear_static_mods(inplace=True)

    return annotation


def condense_to_peptidoform(annotation: ProFormaAnnotation, inplace: bool = True) -> ProFormaAnnotation:
    """
    Condense all modifications into peptidoform notation
    PEP[+100]TI[+20]DE-[30]

    {+100|info:P}{+20|info:T}{-30|info:Nterm}PEPTIDE
    """

    if inplace is False:
        return condense_to_peptidoform(annotation.copy(), inplace=True)

    # pop intervals

    # Collect all modifications as strings with their counts
    if annotation.has_internal_mods:
        internal_mods = annotation.pop_internal_mods()
        for _, mods in internal_mods.items():
            annotation.extend_unknown_mods(mods)

    if annotation.has_intervals:
        intervals = annotation.pop_intervals()

        for interval in intervals:
            if not interval.has_mods:
                continue
            annotation.extend_unknown_mods(interval.mods)

    return annotation


def count_residues(annotation: ProFormaAnnotation, include_mods: bool = True) -> dict[str, int]:
    """
    Count the occurrences of each residue in the sequence.

    Args:
        annotation: The ProFormaAnnotation to analyze
        include_mods: Whether to include modifications in the count

    Returns:
        Dictionary mapping residue strings to their counts
    """

    if not include_mods:
        return dict(Counter(annotation.sequence))

    split_annotations = annotation.split()
    return dict(Counter(a.serialize() for a in split_annotations))


def percent_residues(
    annotation: ProFormaAnnotation,
    include_mods: bool = True,
) -> dict[str, float]:
    """
    Calculate the percentage of each residue in the sequence.

    Args:
        annotation: The ProFormaAnnotation to analyze
        include_mods: Whether to include modifications in the calculation
        precision: Number of decimal places to round to

    Returns:
        Dictionary mapping residue strings to their percentages
    """
    residue_counts = count_residues(annotation, include_mods=include_mods)
    total_count = sum(residue_counts.values())

    if total_count == 0:
        return {}

    percentages = {residue: (count / total_count) * 100 for residue, count in residue_counts.items()}

    return percentages


def is_subsequence(
    annotation: ProFormaAnnotation,
    other: ProFormaAnnotation,
    ignore_mods: bool = False,
    ignore_intervals: bool = True,
) -> bool:
    """
    Check if the annotation is a subsequence of another annotation.

    Args:
        annotation: The annotation to search for
        other: The annotation to search in
        ignore_mods: If True, only compare sequences (ignore modifications)
        ignore_intervals: If True, ignore interval modifications

    Returns:
        True if annotation is a subsequence of other
    """
    if annotation.sequence not in other.sequence:
        return False

    # Loop over all starting indexes where the sequence is found
    for match in re.finditer(annotation.sequence, other.sequence):
        start = match.start()

        # Check if all modifications are also a subsequence
        sliced_annot = other.slice(start, start + len(annotation.sequence), inplace=False)

        if ignore_mods and sliced_annot.sequence == annotation.sequence:
            return True

        if ignore_intervals:
            sliced_clean = _remove_intervals_copy(sliced_annot)
            annot_clean = _remove_intervals_copy(annotation)
            if sliced_clean == annot_clean:
                return True

        if sliced_annot == annotation:
            return True

    return False


def find_indices(
    annotation: ProFormaAnnotation,
    other: ProFormaAnnotation,
    ignore_mods: bool = False,
    ignore_intervals: bool = True,
) -> list[int]:
    """
    Find all occurrences of the annotation in another annotation.

    Args:
        annotation: The annotation to search for
        other: The annotation to search in
        ignore_mods: If True, only compare sequences
        ignore_intervals: If True, ignore interval modifications

    Returns:
        List of starting indices where annotation occurs in other

    Raises:
        TypeError: If other is not a ProFormaAnnotation
    """
    if not isinstance(other, type(annotation)):
        raise TypeError(f"other must be a {type(annotation).__name__}, got {type(other).__name__}")

    if not annotation.sequence or not other.sequence:
        return []

    indices: list[int] = []
    for match in re.finditer(annotation.sequence, other.sequence):
        start = match.start()
        sliced_other = other.slice(start, start + len(annotation.sequence), inplace=False)

        if is_subsequence(annotation, sliced_other, ignore_mods, ignore_intervals):
            indices.append(start)

    return indices


def _remove_intervals_copy(annotation: ProFormaAnnotation) -> ProFormaAnnotation:
    """Create a copy of annotation with intervals removed"""
    copy_annot = annotation.copy()
    copy_annot.intervals = None
    return copy_annot



from peptacular.annotation.annotation import convert_moddict_input


def test_convert_mod_input_dict():
    assert convert_moddict_input({"Oxidation": 1}) == {"Oxidation": 1}
    assert convert_moddict_input({"Oxidation": 2, "Phospho": 1}) == {
        "Oxidation": 2,
        "Phospho": 1,
    }
    # Test non-string keys
    assert convert_moddict_input({1: 1}) == {"1": 1}


def test_convert_mod_input_str():
    assert convert_moddict_input("Oxidation") == {"Oxidation": 1}
    assert convert_moddict_input("Phospho") == {"Phospho": 1}
    assert convert_moddict_input("") == {"": 1}


def test_convert_mod_input_number():
    assert convert_moddict_input(1) == {"+1": 1}
    assert convert_moddict_input(-1) == {"-1": 1}
    assert convert_moddict_input(1.5) == {"+1.5": 1}
    assert convert_moddict_input(-1.5) == {"-1.5": 1}
    assert convert_moddict_input(0) == {"+0": 1}


def test_convert_mod_input_iterable():
    assert convert_moddict_input(["Oxidation", "Oxidation"]) == {"Oxidation": 2}
    assert convert_moddict_input(["Oxidation", "Phospho"]) == {
        "Oxidation": 1,
        "Phospho": 1,
    }
    assert convert_moddict_input((1, 2, 1)) == {"1": 2, "2": 1}


def test_convert_mod_input_empty():
    assert convert_moddict_input({}) == {}
    assert convert_moddict_input([]) == {}


def test_convert_mod_input_none():
    assert convert_moddict_input(None) == {}


def test_convert_mod_input_other():
    # Test with an object that doesn't match any criteria
    class SomeObj:
        pass

    assert convert_moddict_input(SomeObj()) == {}



import unittest

import peptacular as pt


class TestCoverage(unittest.TestCase):
    def test_coverage_basic_exact_match(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [pt.ProFormaAnnotation.parse("PEPTIDE")]
        result = target.coverage(annotations)

        self.assertEqual(result, [1, 1, 1, 1, 1, 1, 1])

    def test_coverage_basic_no_match(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [pt.ProFormaAnnotation.parse("DIFFERENT")]
        result = target.coverage(annotations)

        self.assertEqual(result, [0, 0, 0, 0, 0, 0, 0])

    def test_coverage_subsequence_match(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [pt.ProFormaAnnotation.parse("PEP")]
        result = target.coverage(annotations)

        self.assertEqual(result, [1, 1, 1, 0, 0, 0, 0])

    def test_coverage_multiple_subsequences(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [
            pt.ProFormaAnnotation.parse("PEP"),
            pt.ProFormaAnnotation.parse("TIDE"),
        ]
        result = target.coverage(annotations)

        self.assertEqual(result, [1, 1, 1, 1, 1, 1, 1])

    def test_coverage_overlapping_subsequences(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [
            pt.ProFormaAnnotation.parse("PEPT"),
            pt.ProFormaAnnotation.parse("TIDE"),
        ]
        result = target.coverage(annotations)

        self.assertEqual(result, [1, 1, 1, 1, 1, 1, 1])

    def test_coverage_accumulate_true(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [
            pt.ProFormaAnnotation.parse("PEP"),
            pt.ProFormaAnnotation.parse("PEP"),
        ]  # Same subsequence twice
        result = target.coverage(annotations, accumulate=True)

        self.assertEqual(result, [2, 2, 2, 0, 0, 0, 0])

    def test_coverage_accumulate_false(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [
            pt.ProFormaAnnotation.parse("PEP"),
            pt.ProFormaAnnotation.parse("PEP"),
        ]  # Same subsequence twice
        result = target.coverage(annotations, accumulate=False)

        self.assertEqual(result, [1, 1, 1, 0, 0, 0, 0])

    def test_coverage_ignore_mods_true(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [pt.ProFormaAnnotation.parse("PE[Phospho]PTIDE")]
        result = target.coverage(annotations, ignore_mods=True)

        self.assertEqual(result, [1, 1, 1, 1, 1, 1, 1])

    def test_coverage_ignore_mods_false(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [pt.ProFormaAnnotation.parse("PE[Phospho]PTIDE")]
        result = target.coverage(annotations, ignore_mods=False)

        self.assertEqual(result, [0, 0, 0, 0, 0, 0, 0])

    def test_coverage_with_ambiguous_intervals(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        # Create annotation with ambiguous interval
        annotation = pt.ProFormaAnnotation.parse("P(?EP)TIDE")

        result = target.coverage([annotation], ignore_ambiguity=False)

        # Positions 1 and 2 should have 0 coverage due to ambiguity
        self.assertEqual(result, [1, 0, 0, 1, 1, 1, 1])

    def test_coverage_ignore_ambiguity_true(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        # Create annotation with ambiguous interval
        annotation = pt.ProFormaAnnotation.parse("P(?EP)TIDE")

        result = target.coverage([annotation], ignore_ambiguity=True)

        # All positions should have coverage when ignoring ambiguity
        self.assertEqual(result, [1, 1, 1, 1, 1, 1, 1])

    def test_coverage_empty_annotations_list(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        result = target.coverage([])

        self.assertEqual(result, [0, 0, 0, 0, 0, 0, 0])

    def test_coverage_partial_overlaps(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [
            pt.ProFormaAnnotation.parse("EPT"),
            pt.ProFormaAnnotation.parse("TID"),
        ]
        result = target.coverage(annotations)

        self.assertEqual(result, [0, 1, 1, 1, 1, 1, 0])

    def test_coverage_multiple_overlapping_with_accumulate(self):
        target = pt.ProFormaAnnotation.parse("PEPTIDE")
        annotations = [
            pt.ProFormaAnnotation.parse("PEPT"),
            pt.ProFormaAnnotation.parse("EPT"),
            pt.ProFormaAnnotation.parse("PT"),
        ]
        result = target.coverage(annotations, accumulate=True)

        # P=1, E=2, P=3, T=2, I=0, D=0, E=0
        self.assertEqual(result, [1, 2, 3, 3, 0, 0, 0])

    def test_coverage_with_modifications_in_target(self):
        target = pt.ProFormaAnnotation.parse("PE[Phospho]PTIDE")
        annotations = [pt.ProFormaAnnotation.parse("PEPTIDE")]
        result = target.coverage(annotations, ignore_mods=True)

        self.assertEqual(result, [1, 1, 1, 1, 1, 1, 1])


if __name__ == "__main__":
    unittest.main()



import re
import warnings
from collections.abc import Generator


def get_regex_match_indices(input_str: str, regex_str: str | re.Pattern[str], offset: int = 0) -> Generator[int, None, None]:
    """
    Identify the starting indexes of occurrences of a given regex pattern within a string.

    :param input_str: The sequence in which to search.
    :type input_str: str
    :param regex_str: The regex pattern to search for.
    :type regex_str: str
    :param offset: An optional offset to add to each identified index. Default is 0.
    :type offset: int
    :return: A list of starting indexes where the regex pattern is found in the sequence.
    :rtype: List[int]

    .. code-block:: python

        >>> list(get_regex_match_indices("PEPTIDE", "(?=P)"))
        [0, 2]

        >>> list(get_regex_match_indices("PEPTIDE", re.compile("(?=E)")))
        [1, 6]

        >>> list(get_regex_match_indices("PEPTIDE", '(?=E)'))
        [1, 6]

        # More complex regex
        >>> list(get_regex_match_indices("PEPTIDE", "(?=P[ST])"))
        [2]

        >>> list(get_regex_match_indices("PPPP", "(?=PP)"))
        [0, 1, 2]

        >>> list(get_regex_match_indices("PEPCTIDE", "(?=C)"))
        [3]

        >>> list(get_regex_match_indices("PEPCTIDE", "(?<=P)(?=E)"))
        [1]

        >>> list(get_regex_match_indices("PEPCTIDE", "(?<=C)"))
        [4]


    """

    if not isinstance(regex_str, re.Pattern):
        regex_pattern = re.compile(regex_str)
    else:
        regex_pattern = regex_str

    # Use overlapping search
    pos = 0
    while pos < len(input_str):
        match = regex_pattern.search(input_str, pos)

        if match is None:
            break

        if match.start() != match.end():
            warnings.warn(
                message="The regex pattern has a match with a none zero length. Using start index + 1 for the match.",
                category=UserWarning,
                stacklevel=2,
            )
            yield match.start() + offset + 1
        else:
            yield match.start() + offset

        # Advance by 1 to find overlapping matches
        pos = match.start() + 1


def get_regex_match_range(input_str: str, regex_str: str | re.Pattern[str], offset: int = 0) -> list[tuple[int, int]]:
    """
    Identify the starting indexes of occurrences of a given regex pattern within a string.

    :param input_str: The sequence in which to search.
    :type input_str: str
    :param regex_str: The regex pattern to search for.
    :type regex_str: str
    :param offset: An optional offset to add to each identified index. Default is 0.
    :type offset: int
    :return: A list of starting indexes where the regex pattern is found in the sequence.
    :rtype: List[int]

    .. code-block:: python

        >>> get_regex_match_range("PEPTIDE", "P")
        [(0, 1), (2, 3)]

        >>> get_regex_match_range("PEPTIDE", re.compile("E"))
        [(1, 2), (6, 7)]

        # More complex regex
        >>> get_regex_match_range("PEPTIDE", "P[ST]")
        [(2, 4)]

        >>> get_regex_match_range("PPPP", "PP")
        [(0, 2), (1, 3), (2, 4)]

    """

    if not isinstance(regex_str, re.Pattern):
        regex_pattern = re.compile(regex_str)
    else:
        regex_pattern = regex_str

    # Use overlapping search
    matches: list[tuple[int, int]] = []
    pos = 0
    while pos < len(input_str):
        match = regex_pattern.search(input_str, pos)

        if match is None:
            break

        matches.append((match.start() + offset, match.end() + offset))

        # Advance by 1 to find overlapping matches
        pos = match.start() + 1

    return matches



import peptacular as pt

if __name__ == "__main__":
    # Parse a ProForma sequence
    annot = pt.parse("PEPT[Phospho]IDE-[Acetyl]")

    # Calculate mass and m/z
    mass = annot.mass()
    mz = annot.mz(charge=2)

    print(annot.set_charge(2).set_peptide_name("Peptacular").serialize())


from collections.abc import Sequence
from typing import Any, overload

from ..annotation import (
    ProFormaAnnotation,
)
from ..constants import parallelMethod, parallelMethodLiteral
from .parallel import parallel_apply_internal
from .util import get_annotation_input


def _parse_chimeric_single(s: str, validate: bool = False) -> list[ProFormaAnnotation]:
    return list(ProFormaAnnotation.parse_chimeric(s, validate=validate))


@overload
def parse_chimeric(
    s: str,
    validate: bool = False,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
    reuse_pool: bool = True,
) -> list[ProFormaAnnotation]: ...


@overload
def parse_chimeric(
    s: Sequence[str],
    validate: bool = False,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
    reuse_pool: bool = True,
) -> list[list[ProFormaAnnotation]]: ...


def parse_chimeric(
    s: str | Sequence[str],
    validate: bool = False,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
    reuse_pool: bool = True,
) -> list[ProFormaAnnotation] | list[list[ProFormaAnnotation]]:
    """Parse a chimeric ProForma string or list of strings into lists of ProFormaAnnotation objects."""
    if isinstance(s, Sequence) and not isinstance(s, str):
        return parallel_apply_internal(
            _parse_chimeric_single,
            s,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
            validate=validate,
            reuse_pool=reuse_pool,
        )
    else:
        return _parse_chimeric_single(s, validate=validate)


def _serialize_chimeric_single(
    sequence: Sequence[ProFormaAnnotation | str],
) -> str:
    annots = [get_annotation_input(seq, copy=True) for seq in sequence]

    # first annot will have comound anme and global mods (iso and static)

    # ensure all annots share the same compound name and global mods
    compound_names = {annot.compound_name for annot in annots}
    if len(compound_names) > 1:
        raise ValueError("All annotations in a chimeric sequence must share the same compound name.")

    static_mods = {annot.static_mods for annot in annots}

    if len(static_mods) > 1:
        raise ValueError("All annotations in a chimeric sequence must share the same static modifications.")

    isotope_mods = {annot.isotope_mods for annot in annots}

    if len(isotope_mods) > 1:
        raise ValueError("All annotations in a chimeric sequence must share the same isotopic modifications.")

    for i, annot in enumerate(annots):
        if i == 0:
            continue
        annot.compound_name = None
        annot.static_mods = None
        annot.isotope_mods = None

    return "+".join(annot.serialize() for annot in annots)


@overload
def serialize_chimeric(
    sequence: Sequence[ProFormaAnnotation | str],
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> str: ...


@overload
def serialize_chimeric(
    sequence: Sequence[Sequence[ProFormaAnnotation | str]],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[str]: ...


def serialize_chimeric(
    sequence: Sequence[ProFormaAnnotation | str] | Sequence[Sequence[ProFormaAnnotation | str]],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> str | list[str]:
    """Serialize a chimeric peptide sequence or list of sequences to ProForma string format."""
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and all(isinstance(seq, Sequence) and not isinstance(seq, str) for seq in sequence):
        return parallel_apply_internal(
            _serialize_chimeric_single,
            sequence,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _serialize_chimeric_single(sequence)  # type: ignore[arg-type]


def _parse_single(s: str, validate: bool = False) -> ProFormaAnnotation:
    return ProFormaAnnotation.parse(s, validate=validate)


@overload
def parse(
    s: str,
    validate: bool = False,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
    reuse_pool: bool = True,
) -> ProFormaAnnotation: ...


@overload
def parse(
    s: Sequence[str],
    validate: bool = False,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
    reuse_pool: bool = True,
) -> list[ProFormaAnnotation]: ...


def parse(
    s: str | Sequence[str],
    validate: bool = False,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
    reuse_pool: bool = True,
) -> ProFormaAnnotation | list[ProFormaAnnotation]:
    """Parse a ProForma string or list of strings into ProFormaAnnotation object(s)."""
    if isinstance(s, Sequence) and not isinstance(s, str):
        return parallel_apply_internal(
            _parse_single,
            s,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
            validate=validate,
            reuse_pool=reuse_pool,
        )
    else:
        return _parse_single(s, validate=validate)


def _serialize_single(
    sequence: str | ProFormaAnnotation,
) -> str:
    return get_annotation_input(sequence, copy=False).serialize()


@overload
def serialize(
    sequence: str | ProFormaAnnotation,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> str: ...


@overload
def serialize(
    sequence: Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[str]: ...


def serialize(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> str | list[str]:
    """Serialize a peptide sequence or list of sequences to ProForma string format."""
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _serialize_single,
            sequence,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _serialize_single(sequence)


def _sequence_length_single(sequence: str | ProFormaAnnotation) -> int:
    return len(get_annotation_input(sequence, copy=False))


@overload
def sequence_length(
    sequence: str | ProFormaAnnotation,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> int: ...


@overload
def sequence_length(
    sequence: Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[int]: ...


def sequence_length(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> int | list[int]:
    """Compute the length of the peptide sequence based on the unmodified sequence."""
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _sequence_length_single,
            sequence,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _sequence_length_single(sequence)


def _is_ambiguous_single(sequence: str | ProFormaAnnotation) -> bool:
    return get_annotation_input(sequence, copy=False).has_sequence_ambiguity


@overload
def is_ambiguous(
    sequence: str | ProFormaAnnotation,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> bool: ...


@overload
def is_ambiguous(
    sequence: Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[bool]: ...


def is_ambiguous(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> bool | list[bool]:
    """Check if the sequence contains ambiguous amino acids."""
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _is_ambiguous_single,
            sequence,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _is_ambiguous_single(sequence)


def _is_modified_single(sequence: str | ProFormaAnnotation) -> bool:
    return get_annotation_input(sequence, copy=False).has_mods()


@overload
def is_modified(
    sequence: str | ProFormaAnnotation,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> bool: ...


@overload
def is_modified(
    sequence: Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[bool]: ...


def is_modified(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> bool | list[bool]:
    """Check if the sequence contains any modifications."""
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _is_modified_single,
            sequence,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _is_modified_single(sequence)


def _count_residues_single(sequence: str | ProFormaAnnotation, include_mods: bool = True) -> dict[str, int]:
    return get_annotation_input(sequence, copy=False).condense_static_mods(inplace=True).count_residues(include_mods=include_mods)


@overload
def count_residues(
    sequence: str | ProFormaAnnotation,
    include_mods: bool = True,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> dict[str, int]: ...


@overload
def count_residues(
    sequence: Sequence[str | ProFormaAnnotation],
    include_mods: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[dict[str, int]]: ...


def count_residues(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    include_mods: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> dict[str, int] | list[dict[str, int]]:
    """
    Counts the occurrences of each amino acid in the input sequence.

    include_mods: If True, modified residues are counted as distinct entities. If False,
                  only unmodified residues are counted.

    .. code-block:: python

        # Single sequence
        >>> count_residues('PEPTIDE')
        {'P': 2, 'E': 2, 'T': 1, 'I': 1, 'D': 1}

        # Single sequence
        >>> count_residues('PEP[Oxidation]TIDE-[+30]')
        {'P': 1, 'E': 1, 'P[Oxidation]': 1, 'T': 1, 'I': 1, 'D': 1, 'E-[+30]': 1}

    """
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _count_residues_single,
            sequence,
            include_mods=include_mods,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _count_residues_single(sequence, include_mods=include_mods)


def _percent_residues_single(
    sequence: str | ProFormaAnnotation,
    include_mods: bool = True,
) -> dict[str, float]:
    return get_annotation_input(sequence, copy=False).condense_static_mods(inplace=True).percent_residues(include_mods=include_mods)


@overload
def percent_residues(
    sequence: str | ProFormaAnnotation,
    include_mods: bool = True,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> dict[str, float]: ...


@overload
def percent_residues(
    sequence: Sequence[str | ProFormaAnnotation],
    include_mods: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[dict[str, float]]: ...


def percent_residues(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    include_mods: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> dict[str, float] | list[dict[str, float]]:
    """
    Calculates the percentage of each amino acid in the input sequence.

    include_mods: If True, modified residues are counted as distinct entities. If False,
                  only unmodified residues are counted.

    .. code-block:: python

        # Single sequence
        >>> d = percent_residues('PEPTIDE')
        >>> dict(map(lambda item: (item[0], round(item[1], 2)), d.items()))
        {'P': 28.57, 'E': 28.57, 'T': 14.29, 'I': 14.29, 'D': 14.29}

        # Single sequence with modification
        >>> d = percent_residues('PEP[Oxidation]TIDE-[+30]')
        >>> dict(map(lambda item: (item[0], round(item[1], 2)), d.items()))
        {'P': 14.29, 'E': 14.29, 'P[Oxidation]': 14.29, 'T': 14.29, 'I': 14.29, 'D': 14.29, 'E-[+30]': 14.29}

    """
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _percent_residues_single,
            sequence,
            include_mods=include_mods,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _percent_residues_single(sequence, include_mods=include_mods)


def annotate_ambiguity(
    sequence: str | ProFormaAnnotation,
    forward_coverage: list[int],
    reverse_coverage: list[int],
    mass_shift: Any | None = None,
    add_mods_to_intervals: bool = False,
    sort_mods: bool = True,
    condense_to_xnotation: bool = False,
) -> str:
    """
    This function identifies regions in the sequence where there is insufficient fragment ion
    coverage and marks them as ambiguous using ProForma notation with parentheses.
    If a mass shift is provided, it will be added to the appropriate location.

    forward_coverage: Binary list indicating which positions have forward ion coverage (1) or not (0).
    reverse_coverage: Binary list indicating which positions have reverse ion coverage (1) or not (0).
    mass_shift: An optional mass shift to be added to the sequence at the appropriate position.
    add_mods_to_intervals: Whether to add modifications to interval annotations.
    sort_mods: Whether to sort modifications.
    condense_to_xnotation: Whether to condense ambiguity to X notation.

    .. code-block:: python

        # Add ambiguity intervals based on fragment ion coverage
        >>> annotate_ambiguity('PEPTIDE', [0,1,1,1,0,0,0], [0,0,0,0,0,1,0])
        '(?PE)PTI(?DE)'

        # With a phosphorylation mass shift (note the '+' sign)
        >>> annotate_ambiguity('PEPTIDE', [1,1,1,0,0,0,0], [0,0,0,0,1,1,1], 79.966)
        'PEPT[+79.966]IDE'

        # Handling existing modifications
        >>> annotate_ambiguity('P[+10]EPTIDE', [1,1,1,0,0,0,0], [0,0,0,0,0,1,1])
        'P[+10]EP(?TI)DE'

        # When mass shift can't be localized to a specific residue
        >>> annotate_ambiguity('PEPTIDE', [0,1,1,0,0,0,0], [0,0,0,0,0,1,0], 120)
        '(?PE)P(?TI)[+120](?DE)'

        # When mass shift is completely unlocalized, it becomes a labile modification
        >>> annotate_ambiguity('PEPTIDE', [0,1,1,1,1,0,0], [0,0,1,1,1,1,0], 120)
        '{+120}(?PE)PTI(?DE)'

        # Complex example with multiple intervals
        >>> for_ions = list(map(int, '00011101001000000000000000000000000000'))
        >>> rev_ions = list(map(int, '00000000000110000000101111111111010100'))
        >>> annotate_ambiguity('SSGSIASSYVQWYQQRPGSAPTTVIYEDDERPSGVPDR', for_ions, rev_ions, 120)
        '(?SSGS)IA(?SS)(?YVQ)W[+120](?YQQRPGSA)(?PT)TVIYEDDER(?PS)(?GV)(?PDR)'
    """
    annot = get_annotation_input(sequence=sequence, copy=True).annotate_ambiguity(
        forward_coverage=forward_coverage,
        reverse_coverage=reverse_coverage,
        mass_shift=mass_shift,
        add_mods_to_intervals=add_mods_to_intervals,
        sort_mods=sort_mods,
        inplace=True,
    )

    if condense_to_xnotation:
        annot.condense_ambiguity_to_xnotation(inplace=True)

    return annot.serialize()


def _validate_single(
    sequence: str | ProFormaAnnotation,
) -> bool:
    try:
        get_annotation_input(sequence, copy=False).validate_annotation()
    except ValueError:
        return False
    return True


@overload
def validate(
    sequence: str | ProFormaAnnotation,
    n_workers: None = None,
    chunksize: None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> bool: ...


@overload
def validate(
    sequence: Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[bool]: ...


def validate(
    sequence: str | ProFormaAnnotation | Sequence[str | ProFormaAnnotation],
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> bool | list[bool]:
    """
    Checks if the input sequence is a valid ProForma sequence.
    """
    if isinstance(sequence, Sequence) and not isinstance(sequence, str) and not isinstance(sequence, ProFormaAnnotation):
        return parallel_apply_internal(
            _validate_single,
            sequence,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )
    else:
        return _validate_single(sequence)


def _generate_random_single(
    _item: Any = None,  # Dummy parameter for parallel processing
    min_length: int = 6,
    max_length: int = 20,
    mod_probability: float = 0.05,
    include_internal_mods: bool = True,
    include_nterm_mods: bool = True,
    include_cterm_mods: bool = True,
    include_labile_mods: bool = True,
    include_unknown_mods: bool = True,
    include_isotopic_mods: bool = True,
    include_static_mods: bool = True,
    include_intervals: bool = True,
    include_charge: bool = True,
    require_composition: bool = True,
) -> ProFormaAnnotation:
    return ProFormaAnnotation.random(
        min_length=min_length,
        max_length=max_length,
        mod_probability=mod_probability,
        include_internal_mods=include_internal_mods,
        include_nterm_mods=include_nterm_mods,
        include_cterm_mods=include_cterm_mods,
        include_labile_mods=include_labile_mods,
        include_unknown_mods=include_unknown_mods,
        include_isotopic_mods=include_isotopic_mods,
        include_static_mods=include_static_mods,
        include_intervals=include_intervals,
        include_charge=include_charge,
        require_composition=require_composition,
    )


@overload
def generate_random(
    count: None = None,
    min_length: int = 6,
    max_length: int = 20,
    mod_probability: float = 0.05,
    include_internal_mods: bool = True,
    include_nterm_mods: bool = True,
    include_cterm_mods: bool = True,
    include_labile_mods: bool = True,
    include_unknown_mods: bool = True,
    include_isotopic_mods: bool = True,
    include_static_mods: bool = True,
    include_intervals: bool = True,
    include_charge: bool = True,
    require_composition: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> ProFormaAnnotation: ...


@overload
def generate_random(
    count: int,
    min_length: int = 6,
    max_length: int = 20,
    mod_probability: float = 0.05,
    include_internal_mods: bool = True,
    include_nterm_mods: bool = True,
    include_cterm_mods: bool = True,
    include_labile_mods: bool = True,
    include_unknown_mods: bool = True,
    include_isotopic_mods: bool = True,
    include_static_mods: bool = True,
    include_intervals: bool = True,
    include_charge: bool = True,
    require_composition: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> list[ProFormaAnnotation]: ...


def generate_random(
    count: int | None = None,
    min_length: int = 6,
    max_length: int = 20,
    mod_probability: float = 0.05,
    include_internal_mods: bool = True,
    include_nterm_mods: bool = True,
    include_cterm_mods: bool = True,
    include_labile_mods: bool = True,
    include_unknown_mods: bool = True,
    include_isotopic_mods: bool = True,
    include_static_mods: bool = True,
    include_intervals: bool = True,
    include_charge: bool = True,
    require_composition: bool = True,
    n_workers: int | None = None,
    chunksize: int | None = None,
    method: parallelMethod | parallelMethodLiteral | None = None,
) -> ProFormaAnnotation | list[ProFormaAnnotation]:
    """Generate random ProForma annotation(s) with configurable features.

    Args:
        count: Number of random sequences to generate. If None, generates a single sequence.
        min_length: Minimum sequence length
        max_length: Maximum sequence length
        mod_probability: Probability of adding modifications (0.0 to 1.0)
        include_internal_mods: Whether to generate internal modifications
        include_nterm_mods: Whether to generate N-terminal modifications
        include_cterm_mods: Whether to generate C-terminal modifications
        include_labile_mods: Whether to generate labile modifications
        include_unknown_mods: Whether to generate unknown position modifications
        include_isotopic_mods: Whether to generate isotopic modifications
        include_static_mods: Whether to generate static modifications
        include_intervals: Whether to generate intervals
        include_charge: Whether to generate charge state or adduct
        require_composition: If True, only modifications with composition are allowed (no mass-only)
        n_workers: Number of parallel workers (only used when count > 1)
        chunksize: Size of chunks for parallel processing
        method: Parallel processing method ('process', 'thread', or 'sequential')

    Returns:
        A single ProFormaAnnotation if count is None, otherwise a list of ProFormaAnnotations

    .. code-block:: python

        # Generate a single random sequence
        >>> seq = generate_random()
        >>> isinstance(seq, ProFormaAnnotation)
        True

        # Generate multiple random sequences
        >>> seqs = generate_random(count=10)
        >>> len(seqs)
        10

        # Generate without modifications
        >>> seq = generate_random(mod_probability=0.0)

        # Generate with only internal modifications
        >>> seq = generate_random(
        ...     include_nterm_mods=False,
        ...     include_cterm_mods=False,
        ...     include_labile_mods=False
        ... )
    """
    if count is None:
        return _generate_random_single(
            min_length=min_length,
            max_length=max_length,
            mod_probability=mod_probability,
            include_internal_mods=include_internal_mods,
            include_nterm_mods=include_nterm_mods,
            include_cterm_mods=include_cterm_mods,
            include_labile_mods=include_labile_mods,
            include_unknown_mods=include_unknown_mods,
            include_isotopic_mods=include_isotopic_mods,
            include_static_mods=include_static_mods,
            include_intervals=include_intervals,
            include_charge=include_charge,
            require_composition=require_composition,
        )
    else:
        # Generate count number of items
        items = [None] * count
        return parallel_apply_internal(
            _generate_random_single,
            items,
            min_length=min_length,
            max_length=max_length,
            mod_probability=mod_probability,
            include_internal_mods=include_internal_mods,
            include_nterm_mods=include_nterm_mods,
            include_cterm_mods=include_cterm_mods,
            include_labile_mods=include_labile_mods,
            include_unknown_mods=include_unknown_mods,
            include_isotopic_mods=include_isotopic_mods,
            include_static_mods=include_static_mods,
            include_intervals=include_intervals,
            include_charge=include_charge,
            require_composition=require_composition,
            n_workers=n_workers,
            chunksize=chunksize,
            method=method,
        )



"""Protein property calculation module."""

from .data import (
    COMPOSITION_SCALES,
    FLIXIBILITY_SCALES,
    HPLC_SCALES,
    HYDROPHILICITY_SCALES,
    HYDROPHOBICITY_SCALES,
    PHYSICAL_PROPERTY_SCALES,
    POLARITY_SCALES,
    PROPERTY_SCALES,
    SURFACE_ACCESSIBILITY_SCALES,
    BetaStrandScale,
    ChargeScale,
    CompositionScale,
    HPLCScale,
    HydrophobicityScale,
    PhysicalPropertyScale,
    PolarityScale,
    PropertyScale,
    SecondaryStructureMethod,
    SecondaryStructureScale,
    SecondaryStructureType,
    SurfaceAccessibilityScale,
)
from .prop import AnnotationProperties
from .types import (
    AggregationMethod,
    AggregationMethodLiteral,
    MissingAAHandling,
    MissingAAHandlingLiteral,
    WeightingMethods,
    WeightingMethodsLiteral,
)

__all__ = [
    # main class
    "AnnotationProperties",
    # data
    "SecondaryStructureMethod",
    "SecondaryStructureType",
    "PropertyScale",
    "HydrophobicityScale",
    "SecondaryStructureScale",
    "SurfaceAccessibilityScale",
    "ChargeScale",
    "PolarityScale",
    "HPLCScale",
    "BetaStrandScale",
    "PhysicalPropertyScale",
    "CompositionScale",
    "PROPERTY_SCALES",
    "HYDROPHOBICITY_SCALES",
    "SURFACE_ACCESSIBILITY_SCALES",
    "HPLC_SCALES",
    "HYDROPHILICITY_SCALES",
    "FLIXIBILITY_SCALES",
    "POLARITY_SCALES",
    "COMPOSITION_SCALES",
    "PHYSICAL_PROPERTY_SCALES",
    # types
    "MissingAAHandling",
    "MissingAAHandlingLiteral",
    "AggregationMethod",
    "AggregationMethodLiteral",
    "WeightingMethods",
    "WeightingMethodsLiteral",
]



"""
Tests for the from_string class methods on NamedTuples.
Demonstrates round-trip conversion: string -> object -> string
"""

import pytest

import peptacular as pt


class TestFormulaElementFromString:
    """Test pt.FormulaElement.from_string()"""

    def test_simple_element(self):
        """Test parsing simple element like 'C'"""
        element = pt.FormulaElement.from_string("C")
        assert element.element == pt.Element.C
        assert element.occurance == 1
        assert element.isotope is None
        assert str(element) == "C"

    def test_element_with_count(self):
        """Test parsing element with count like 'H2'"""
        element = pt.FormulaElement.from_string("H2")
        assert element.element == pt.Element.H
        assert element.occurance == 2
        assert element.isotope is None
        assert str(element) == "H2"

    def test_element_with_isotope(self):
        """Test parsing element with isotope like '[13C2]'"""
        element = pt.FormulaElement.from_string("[13C2]")
        assert element.element == pt.Element.C
        assert element.occurance == 2
        assert element.isotope == 13
        assert str(element) == "[13C2]"

    def test_round_trip(self):
        """Test that from_string(str(obj)) == obj"""
        original = pt.FormulaElement(element=pt.Element.N, occurance=3, isotope=15)
        parsed = pt.FormulaElement.from_string(str(original))
        assert parsed == original


class TestChargedFormulaFromString:
    """Test pt.ChargedFormula.from_string()"""

    def test_simple_formula(self):
        """Test parsing simple formula like 'Formula:C2H6'"""
        formula = pt.ChargedFormula.from_string("Formula:C2H6")
        assert len(formula.formula) == 2
        assert formula.charge is None
        assert str(formula) == "Formula:C2H6"

    def test_formula_with_charge(self):
        """Test parsing formula with charge like 'Formula:Na:z+1'"""
        formula = pt.ChargedFormula.from_string("Formula:Na:z+1")
        assert formula.charge == 1
        assert str(formula) == "Formula:Na:z+1"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "Formula:C2H3N:z+2"
        parsed = pt.ChargedFormula.from_string(original_str)
        assert str(parsed) == original_str


class TestTagAccessionFromString:
    """Test pt.TagAccession.from_string()"""

    def test_unimod_full(self):
        """Test parsing UNIMOD accession"""
        tag = pt.TagAccession.from_string("UNIMOD:35")
        assert tag.cv == pt.CV.UNIMOD
        assert tag.accession == "35"
        # Note: str() outputs the CV enum value which is 'UNIMOD' (proper case)
        assert str(tag) == "UNIMOD:35"

    def test_unimod_shorthand(self):
        """Test parsing UNIMOD shorthand"""
        tag = pt.TagAccession.from_string("UNIMOD:35")
        assert tag.cv == pt.CV.UNIMOD
        assert tag.accession == "35"
        # Note: str() always outputs full form with proper CV casing
        assert str(tag) == "UNIMOD:35"

    def test_psi_mod(self):
        """Test parsing PSI-MOD accession"""
        tag = pt.TagAccession.from_string("MOD:00719")
        assert tag.cv == pt.CV.PSI_MOD
        assert tag.accession == "00719"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "XLMOD:02001"
        parsed = pt.TagAccession.from_string(original_str)
        assert str(parsed) == "XLMOD:02001"  # Note: XLMOD normalizes to XL-MOD


class TestTagMassFromString:
    """Test pt.TagMass.from_string()"""

    def test_positive_mass(self):
        """Test parsing positive mass"""
        tag = pt.TagMass.from_string("+15.995")
        assert tag.mass == 15.995
        assert tag.cv is None
        assert str(tag) == "+15.995"

    def test_negative_mass(self):
        """Test parsing negative mass"""
        tag = pt.TagMass.from_string("-18.010")
        assert tag.mass == -18.010
        # Note: Python's float formatting may drop trailing zeros
        assert str(tag) == "-18.01"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "+79.966"
        parsed = pt.TagMass.from_string(original_str)
        assert str(parsed) == original_str


class TestTagNameFromString:
    """Test pt.TagName.from_string()"""

    def test_simple_name(self):
        """Test parsing simple name"""
        tag = pt.TagName.from_string("Oxidation")
        assert tag.name == "Oxidation"
        assert tag.cv is None
        assert str(tag) == "Oxidation"

    def test_name_with_cv(self):
        """Test parsing name with single-letter CV prefix (Rule 4)"""
        tag = pt.TagName.from_string("U:Oxidation")
        assert tag.name == "Oxidation"
        assert tag.cv == pt.CV.UNIMOD
        # Note: str() outputs with proper CV casing
        assert str(tag) == "Oxidation"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "Phospho"
        parsed = pt.TagName.from_string(original_str)
        assert str(parsed) == original_str


class TestTagInfoFromString:
    """Test pt.TagInfo.from_string()"""

    def test_simple_info(self):
        """Test parsing INFO tag"""
        tag = pt.TagInfo.from_string("INFO:probable oxidation")
        assert tag.info == "probable oxidation"
        assert str(tag) == "INFO:probable oxidation"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "INFO:some custom text"
        parsed = pt.TagInfo.from_string(original_str)
        assert str(parsed) == original_str


class TestGlycanComponentFromString:
    """Test pt.GlycanComponent.from_string()"""

    def test_simple_glycan(self):
        """Test parsing simple glycan component"""
        component = pt.GlycanComponent.from_string("Hex")
        assert component.monosaccharide == pt.Monosaccharide.Hex
        assert component.occurance == 1
        assert str(component) == "Hex"

    def test_glycan_with_count(self):
        """Test parsing glycan component with count"""
        component = pt.GlycanComponent.from_string("HexNAc4")
        assert component.monosaccharide == pt.Monosaccharide.HexNAc
        assert component.occurance == 4
        assert str(component) == "HexNAc4"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "NeuAc2"
        parsed = pt.GlycanComponent.from_string(original_str)
        assert str(parsed) == original_str


class TestTagCustomFromString:
    """Test pt.TagCustom.from_string()"""

    def test_custom_tag(self):
        """Test parsing custom tag"""
        tag = pt.TagCustom.from_string("C:MyCustomMod")
        assert tag.name == "MyCustomMod"
        assert str(tag) == "C:MyCustomMod"

    def test_invalid_without_prefix(self):
        """Test that custom tag without C: prefix raises error"""
        with pytest.raises(ValueError):
            pt.TagCustom.from_string("MyCustomMod")


class TestPositionRuleFromString:
    """Test pt.PositionRule.from_string()"""

    def test_anywhere(self):
        """Test parsing amino acid position (anywhere)"""
        rule = pt.PositionRule.from_string("K")
        assert rule.terminal == pt.Terminal.ANYWHERE
        assert rule.amino_acid == pt.AminoAcid.K
        assert str(rule) == "K"

    def test_n_term(self):
        """Test parsing N-terminal position"""
        rule = pt.PositionRule.from_string("N-term")
        assert rule.terminal == pt.Terminal.N_TERM
        assert rule.amino_acid is None
        assert str(rule) == "N-term"

    def test_n_term_with_aa(self):
        """Test parsing N-terminal with specific amino acid"""
        rule = pt.PositionRule.from_string("N-term:K")
        assert rule.terminal == pt.Terminal.N_TERM
        assert rule.amino_acid == pt.AminoAcid.K
        assert str(rule) == "N-term:K"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "C-term:R"
        parsed = pt.PositionRule.from_string(original_str)
        assert str(parsed) == original_str


class TestSequenceElementFromString:
    """Test pt.SequenceElement.from_string()"""

    def test_simple_amino_acid(self):
        """Test parsing simple amino acid"""
        element = pt.SequenceElement.from_string("M")
        assert element.amino_acid == pt.AminoAcid.M
        assert len(element.modifications) == 0
        assert str(element) == "M"

    def test_amino_acid_with_modification(self):
        """Test parsing amino acid with modification"""
        element = pt.SequenceElement.from_string("M[Oxidation]")
        assert element.amino_acid == pt.AminoAcid.M
        assert len(element.modifications) == 1
        assert str(element) == "M[Oxidation]"

    def test_amino_acid_with_multiple_modifications(self):
        """Test parsing amino acid with multiple modifications"""
        element = pt.SequenceElement.from_string("K[UNIMOD:1][+42.010]")
        assert element.amino_acid == pt.AminoAcid.K
        assert len(element.modifications) == 2
        # Note: output format may differ slightly due to normalization

    def test_round_trip_simple(self):
        """Test round-trip conversion for simple case"""
        original_str = "A"
        parsed = pt.SequenceElement.from_string(original_str)
        assert str(parsed) == original_str

    def test_round_trip_with_mod(self):
        """Test round-trip conversion with modification"""
        original_str = "S[Phospho]"
        parsed = pt.SequenceElement.from_string(original_str)
        assert str(parsed) == original_str


class TestIsotopeReplacementFromString:
    """Test pt.IsotopeReplacement.from_string()"""

    def test_deuterium(self):
        """Test parsing deuterium (special case)"""
        iso = pt.IsotopeReplacement.from_string("D")
        assert iso.element == pt.Element.H
        assert iso.isotope == 2
        assert str(iso) == "D"

    def test_carbon_13(self):
        """Test parsing 13C"""
        iso = pt.IsotopeReplacement.from_string("13C")
        assert iso.element == pt.Element.C
        assert iso.isotope == 13
        assert str(iso) == "13C"

    def test_nitrogen_15(self):
        """Test parsing 15N"""
        iso = pt.IsotopeReplacement.from_string("15N")
        assert iso.element == pt.Element.N
        assert iso.isotope == 15
        assert str(iso) == "15N"

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "18O"
        parsed = pt.IsotopeReplacement.from_string(original_str)
        assert str(parsed) == original_str


class TestGlobalChargeCarrierFromString:
    """Test pt.GlobalChargeCarrier.from_string()"""

    def test_simple_carrier(self):
        """Test parsing simple charge carrier"""
        carrier = pt.GlobalChargeCarrier.from_string("Na:z+1")
        assert carrier.occurance == 1.0
        assert carrier.charged_formula.charge == 1
        assert str(carrier) == "Na:z+1"

    def test_carrier_with_occurrence(self):
        """Test parsing charge carrier with occurrence"""
        carrier = pt.GlobalChargeCarrier.from_string("H:z+1^2")
        assert carrier.occurance == 2.0
        assert carrier.charged_formula.charge == 1
        assert str(carrier) == "H:z+1^2"

    def test_carrier_with_fractional_occurrence(self):
        """Test parsing charge carrier with fractional occurrence"""
        with pytest.raises(ValueError):
            carrier = pt.GlobalChargeCarrier.from_string("Na:z+1^1.5")

    def test_round_trip(self):
        """Test round-trip conversion"""
        original_str = "H:z+1^3"
        parsed = pt.GlobalChargeCarrier.from_string(original_str)
        assert str(parsed) == original_str


class TestCaching:
    """Test that from_string methods use caching"""

    def test_same_string_returns_same_object(self):
        """Test that parsing the same string twice returns cached result"""
        tag1 = pt.TagName.from_string("Oxidation")
        tag2 = pt.TagName.from_string("Oxidation")
        # Due to caching, these should be the exact same object
        assert tag1 is tag2

    def test_different_strings_return_different_objects(self):
        """Test that different strings return different objects"""
        tag1 = pt.TagName.from_string("Oxidation")
        tag2 = pt.TagName.from_string("Phospho")
        assert tag1 is not tag2
        assert tag1 != tag2


class TestErrorHandling:
    """Test error handling in from_string methods"""

    def test_invalid_element(self):
        """Test that invalid element raises error"""
        with pytest.raises(ValueError):
            pt.FormulaElement.from_string("Zz")

    def test_invalid_accession(self):
        """Test that invalid accession format raises error"""
        with pytest.raises(ValueError):
            pt.TagAccession.from_string("NotAnAccession")

    def test_invalid_mass(self):
        """Test that invalid mass format raises error"""
        with pytest.raises(ValueError):
            pt.TagMass.from_string("15.995")  # Missing required sign

    def test_invalid_info(self):
        """Test that invalid INFO format raises error"""
        with pytest.raises(ValueError):
            pt.TagInfo.from_string("NotInfo")  # Missing 'INFO:' prefix



import unittest

import peptacular as pt


class TestReverse(unittest.TestCase):
    def test_basic_reverse(self):
        """Test basic sequence reversal"""
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        reversed_annotation = annotation.reverse()

        self.assertEqual(reversed_annotation.serialize(), "EDITPEP")
        self.assertEqual(annotation.serialize(), "PEPTIDE")  # Original unchanged

    def test_basic_reverse_inplace(self):
        """Test in-place sequence reversal"""
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE")
        reversed_annotation = annotation.reverse(inplace=True)

        self.assertEqual(reversed_annotation.serialize(), "EDITPEP")
        self.assertEqual(annotation.serialize(), "EDITPEP")  # Original changed too

    def test_reverse_single_amino_acid(self):
        """Test reversing single amino acid"""
        annotation = pt.ProFormaAnnotation.parse("A")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "A")

    def test_reverse_empty_sequence(self):
        """Test reversing empty sequence"""
        annotation = pt.ProFormaAnnotation.parse("")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "")

    def test_reverse_with_internal_mods(self):
        """Test reversing with internal modifications"""
        annotation = pt.ProFormaAnnotation.parse("PE[Phospho]PTI[Methyl]DE")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "EDI[Methyl]TPE[Phospho]P")

    def test_reverse_with_multiple_internal_mods_same_position(self):
        """Test reversing with multiple modifications on same position"""
        annotation = pt.ProFormaAnnotation.parse("PE[Phospho][Methyl]PTIDE")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "EDITPE[Phospho][Methyl]P")

    def test_reverse_with_intervals(self):
        """Test reversing with intervals"""
        annotation = pt.ProFormaAnnotation.parse("P(EP)[Phospho]TIDE")
        reversed_annotation = annotation.reverse()
        # Original interval was positions 1-3, reversed should be positions 3-5 (7-1-3=3, 7-1-1=5)
        self.assertEqual(reversed_annotation.serialize(), "EDIT(PE)[Phospho]P")

    def test_reverse_with_ambiguous_intervals(self):
        """Test reversing with ambiguous intervals"""
        annotation = pt.ProFormaAnnotation.parse("P(?EPTI)[Phospho]DE")
        reversed_annotation = annotation.reverse()
        # Original interval was positions 1-5, reversed should be positions 1-5 (7-1-5=1, 7-1-1=5)
        self.assertEqual(reversed_annotation.serialize(), "ED(?ITPE)[Phospho]P")

    def test_reverse_with_multiple_intervals(self):
        """Test reversing with multiple intervals"""
        annotation = pt.ProFormaAnnotation.parse("P(EP)[Phospho]TI(DE)[Methyl]")
        reversed_annotation = annotation.reverse()
        # First interval (1,3) becomes (4,6), second interval (4,6) becomes (1,3)
        self.assertEqual(reversed_annotation.serialize(), "(ED)[Methyl]IT(PE)[Phospho]P")

    def test_reverse_with_labile_mods(self):
        """Test reversing with labile modifications"""
        annotation = pt.ProFormaAnnotation.parse("{Glycan}PEPTIDE")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "{Glycan}EDITPEP")

    def test_reverse_with_static_mods(self):
        """Test reversing with static modifications"""
        annotation = pt.ProFormaAnnotation.parse("<57@C>PEPTIDE")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "<57@C>EDITPEP")

    def test_reverse_with_isotope_mods(self):
        """Test reversing with isotope modifications"""
        annotation = pt.ProFormaAnnotation.parse("<15N>PEPTIDE")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "<15N>EDITPEP")

    def test_reverse_with_charge(self):
        """Test reversing with charge state"""
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE/2")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "EDITPEP/2")

    def test_reverse_with_unknown_mod(self):
        """Test reversing with unknown modifications"""
        annotation = pt.ProFormaAnnotation.parse("[Unknown]?PEPTIDE")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "[Unknown]?EDITPEP")

    def test_reverse_with_charge_adducts(self):
        """Test reversing with charge adducts"""
        annotation = pt.ProFormaAnnotation.parse("PEPTIDE/[Na:z+1]")
        reversed_annotation = annotation.reverse()
        self.assertEqual(reversed_annotation.serialize(), "EDITPEP/[Na:z+1]")

    def test_reverse_with_complex_modifications(self):
        """Test reversing with multiple types of modifications"""
        annotation = pt.ProFormaAnnotation.parse("<15N>{Glycan}[Acetyl]-PE[Phospho]PTI[Methyl]DE-[Amide]/2")
        reversed_annotation = annotation.reverse()
        self.assertEqual(
            reversed_annotation.serialize(),
            "<15N>{Glycan}[Acetyl]-EDI[Methyl]TPE[Phospho]P-[Amide]/2",
        )

    def test_reverse_preserves_original_when_not_inplace(self):
        """Test that original annotation is preserved when not in-place"""
        original_annotation = pt.ProFormaAnnotation.parse("PE[Phospho]PTIDE")
        original_serialized = original_annotation.serialize()

        reversed_annotation = original_annotation.reverse()

        self.assertEqual(original_annotation.serialize(), original_serialized)
        self.assertEqual(reversed_annotation.serialize(), "EDITPE[Phospho]P")


if __name__ == "__main__":
    unittest.main()



import peptacular as pt


class TestSequencePropertyMixin:
    """Test the SequencePropertyMixin methods via pt.ProFormaAnnotation"""

    def test_hydrophobicity(self):
        """Test hydrophobicity calculation"""
        # Hydrophobic peptide
        annot = pt.ProFormaAnnotation(sequence="LIVM")
        assert isinstance(annot.prop.hydrophobicity, float)

        # Hydrophilic peptide
        annot = pt.ProFormaAnnotation(sequence="RKDE")
        assert isinstance(annot.prop.hydrophobicity, float)

    def test_pi(self):
        """Test isoelectric point calculation"""
        # Acidic peptide
        annot = pt.ProFormaAnnotation(sequence="DE")
        assert annot.prop.pi < 7.0

        # Basic peptide
        annot = pt.ProFormaAnnotation(sequence="RK")
        assert annot.prop.pi > 7.0

        # Neutral-ish
        annot = pt.ProFormaAnnotation(sequence="A")
        assert 5.0 < annot.prop.pi < 7.0

    def test_charge_at_ph(self):
        """Test charge calculation at specific pH"""
        annot = pt.ProFormaAnnotation(sequence="K")  # Lysine, basic

        # At acidic pH, should be positive
        assert annot.prop.charge_at_ph(pH=2.0) > 0

        # At basic pH, should be neutral or negative (deprotonated)
        assert annot.prop.charge_at_ph(pH=12.0) <= 0

    def test_aromaticity(self):
        """Test aromaticity calculation"""
        # All aromatic
        annot = pt.ProFormaAnnotation(sequence="WFY")
        assert isinstance(annot.prop.aromaticity, float)

        # None aromatic
        annot = pt.ProFormaAnnotation(sequence="A")
        assert isinstance(annot.prop.aromaticity, float)

        # Mixed
        annot = pt.ProFormaAnnotation(sequence="WA")
        assert isinstance(annot.prop.aromaticity, float)

    def test_secondary_structure(self):
        """Test secondary structure propensities"""
        annot = pt.ProFormaAnnotation(sequence="AAAA")

        # Just check that we get values between 0 and 1 (or reasonable percentages)
        # Note: The implementation might return raw scores or normalized values.
        # Based on mixin, it returns values from a scale.

        assert isinstance(annot.prop.alpha_helix_percent, float)
        assert isinstance(annot.prop.beta_sheet_percent, float)
        assert isinstance(annot.prop.beta_turn_percent, float)
        assert isinstance(annot.prop.coil_percent, float)

    def test_calc_property_generic(self):
        """Test generic calc_property method"""
        annot = pt.ProFormaAnnotation(sequence="A")
        # Using a dummy scale
        scale = {"A": 10.0}
        val = annot.prop.calc_property(scale=scale)
        assert val == 10.0

    def test_other_properties(self):
        """Test other properties exist and return floats"""
        annot = pt.ProFormaAnnotation(sequence="ACDEFGHIKLMNPQRSTVWY")

        assert isinstance(annot.prop.flexibility, float)
        assert isinstance(annot.prop.hydrophilicity, float)
        assert isinstance(annot.prop.surface_accessibility, float)
        assert isinstance(annot.prop.polarity, float)
        assert isinstance(annot.prop.mutability, float)
        assert isinstance(annot.prop.codons, float)
        assert isinstance(annot.prop.bulkiness, float)
        assert isinstance(annot.prop.recognition_factors, float)
        assert isinstance(annot.prop.transmembrane_tendency, float)
        assert isinstance(annot.prop.average_buried_area, float)
        assert isinstance(annot.prop.hplc, float)
        assert isinstance(annot.prop.refractivity, float)

    def test_sliding_window(self):
        """Test sliding window feature generation"""
        annot = pt.ProFormaAnnotation(sequence="AAAAA")
        scale = {"A": 1.0}
        # Window size 3, sequence length 5 -> 3 windows?
        # Depends on implementation details (padding etc).
        # generate_sliding_window_features usually returns a list of values.

        features = annot.prop.property_partitions(scale=scale, num_windows=3)
        assert isinstance(features, list)
        assert len(features) > 0
        assert all(isinstance(x, float) for x in features)



import cProfile
import pstats
import random
import time
from io import StringIO

import peptacular as pt


def generate_random_protein(length: int = 500) -> str:
    """Generate a random protein sequence."""
    amino_acids = "ACDEFGHIKLMNPQRSTVWY"
    return "".join(random.choice(amino_acids) for _ in range(length))


def run_digest():
    """Run digest on random proteins."""
    # Generate 500 DIFFERENT proteins instead of 500 copies of the same one
    proteins = [generate_random_protein(1_000) for _ in range(100)]

    start_time = time.time()
    peptides = list(pt.simple_digest(proteins, cleave_on="KR", missed_cleavages=2, semi=False, min_len=6, max_len=50))
    elapsed = time.time() - start_time

    total_peptides = len(peptides)
    print(f"Generated {total_peptides} peptides from {len(proteins)} proteins")
    print(f"Took {elapsed:.2f} seconds ({elapsed / len(proteins) * 1000:.2f} ms per protein)")
    return peptides


def run_digest_single():
    """Run digest on a single protein to profile without multiprocessing."""
    protein = generate_random_protein(1_000)

    start_time = time.time()
    peptides = pt.simple_digest(
        protein,  # Single protein
        cleave_on="KR",
        missed_cleavages=2,
        semi=False,
        min_len=6,
        max_len=50,
    )
    elapsed = time.time() - start_time

    total_peptides = len(list(peptides))
    print(f"Generated {total_peptides} peptides from 1 protein")
    print(f"Took {elapsed * 1000:.2f} ms")
    return peptides


def profile_digest_single():
    """Profile single digest (no multiprocessing overhead)."""
    profiler = cProfile.Profile()
    profiler.enable()
    result = run_digest_single()
    profiler.disable()

    # Print stats
    s = StringIO()
    ps = pstats.Stats(profiler, stream=s).sort_stats("cumulative")
    ps.print_stats(30)  # Top 30 functions
    print(s.getvalue())
    return result


def profile_digest_multi():
    """Profile multi-protein digest (with multiprocessing)."""
    profiler = cProfile.Profile()
    profiler.enable()
    result = run_digest()
    profiler.disable()

    # Print stats
    s = StringIO()
    ps = pstats.Stats(profiler, stream=s).sort_stats("cumulative")
    ps.print_stats(30)  # Top 30 functions
    print(s.getvalue())
    return result


if __name__ == "__main__":
    print("=" * 60)
    print("SINGLE PROTEIN DIGEST (no multiprocessing)")
    print("=" * 60)
    profile_digest_single()

    print("\n" + "=" * 60)
    print("MULTI-PROTEIN DIGEST (with multiprocessing)")
    print("=" * 60)
    # profile_digest_multi()
